import numpy as np
import csv
import cv2
import zebrazoom.videoFormatConversion.zzVideoReading as zzVideoReading
import math
import os
import pickle
import sys
import queue

from zebrazoom.code.extractParameters import extractParameters

from ._base import register_tracking_method
from ._baseZebraZoom import BaseZebraZoomTrackingMethod
from ._eyeTracking import EyeTrackingMixin
from ._getImages import GetImagesMixin
from ._tailTracking import TailTrackingMixin
from ._tailTrackingDifficultBackground import TailTrackingDifficultBackgroundMixin

from zebrazoom.code.vars import getGlobalVariables
globalVariables = getGlobalVariables()


class Tracking(BaseZebraZoomTrackingMethod, TailTrackingDifficultBackgroundMixin, EyeTrackingMixin, GetImagesMixin, TailTrackingMixin):
  def __init__(self, videoPath, wellPositions, hyperparameters):
    self._videoPath = videoPath
    self._hyperparameters = hyperparameters
    self._background = None
    self._wellPositions = wellPositions
    self._videoName = os.path.splitext(os.path.basename(videoPath))[0]
    self._auDessusPerAnimalId = None
    self._firstFrame = self._hyperparameters["firstFrame"] if self._hyperparameters["firstFrameForTracking"] == -1 else self._hyperparameters["firstFrameForTracking"]
    self._lastFrame = self._hyperparameters["lastFrame"]
    self._nbTailPoints = self._hyperparameters["nbTailPoints"]
    self._headPositionFirstFrame = []
    self._tailTipFirstFrame = []
    self._trackingHeadTailAllAnimals = np.zeros((self._hyperparameters["nbAnimalsPerWell"], self._lastFrame-self._firstFrame+1, self._nbTailPoints, 2))
    self._trackingHeadingAllAnimals = np.zeros((self._hyperparameters["nbAnimalsPerWell"], self._lastFrame-self._firstFrame+1))
    if self._hyperparameters["eyeTracking"]:
      self._trackingEyesAllAnimals = np.zeros((self._hyperparameters["nbAnimalsPerWell"], self._lastFrame-self._firstFrame+1, 8))
    else:
      self._trackingEyesAllAnimals = 0

    if not(self._hyperparameters["nbAnimalsPerWell"] > 1 or self._hyperparameters["forceBlobMethodForHeadTracking"]) and not(self._hyperparameters["headEmbeded"]) and (self._hyperparameters["findHeadPositionByUserInput"] == 0) and (self._hyperparameters["takeTheHeadClosestToTheCenter"] == 0):
      self._trackingProbabilityOfGoodDetection = np.zeros((self._hyperparameters["nbAnimalsPerWell"], self._lastFrame-self._firstFrame+1))
    else:
      self._trackingProbabilityOfGoodDetection = 0

    self.useGUI = True

  def _adjustParameters(self, i, initialCurFrame, frame, frame2, widgets):
    return None

  def _addBlackLineToImgSetParameters(self, frame):
    raise ValueError("GUI is required to add black line to image")

  def findTailTipByUserInput(self, frame, frameNumber, wellNumber):
    raise ValueError("GUI is required to select tail tip coordinates, please specify them in advance using the GUI or use --use-gui option")

  def findHeadPositionByUserInput(self, frame, frameNumber, wellNumber):
    raise ValueError("GUI is required to select head coordinates, please specify them in advance using the GUI or use --use-gui option")

  def _getTailTipByFileSaved(self):
    ix = -1
    iy = -1
    with open(self._videoPath+'.csv') as csv_file:
      csv_reader = csv.reader(csv_file, delimiter=',')
      line_count = 0
      for row in csv_reader:
        if len(row):
          ix = row[0]
          iy = row[1]
    return [int(ix),int(iy)]

  def _getHeadPositionByFileSaved(self):
    ix = -1
    iy = -1
    with open(self._videoPath+'HP.csv') as csv_file:
      csv_reader = csv.reader(csv_file, delimiter=',')
      line_count = 0
      for row in csv_reader:
        if len(row):
          ix = row[0]
          iy = row[1]
    return [int(ix),int(iy)]

  def _detectMovementWithRawVideoInsideTracking(self, i, xHead, yHead, initialCurFrame):
    previousFrames   = queue.Queue(self._hyperparameters["frameGapComparision"])
    previousXYCoords = queue.Queue(self._hyperparameters["frameGapComparision"])
    self._auDessusPerAnimalId = [np.zeros((self._lastFrame-self._firstFrame+1, 1)) for _ in range(self._hyperparameters["nbAnimalsPerWell"])]
    halfDiameterRoiBoutDetect = self._hyperparameters["halfDiameterRoiBoutDetect"]
    if previousFrames.full():
      previousFrame   = previousFrames.get()
      curFrame        = initialCurFrame.copy()
      previousXYCoord = previousXYCoords.get()
      curXYCoord      = [xHead, yHead]
      if previousXYCoord[0] < curXYCoord[0]:
        previousFrame = previousFrame[:, (curXYCoord[0]-previousXYCoord[0]):]
      elif previousXYCoord[0] > curXYCoord[0]:
        curFrame      = curFrame[:, (previousXYCoord[0]-curXYCoord[0]):]
      if previousXYCoord[1] < curXYCoord[1]:
        previousFrame = previousFrame[(curXYCoord[1]-previousXYCoord[1]):, :]
      elif previousXYCoord[1] > curXYCoord[1]:
        curFrame      = curFrame[(previousXYCoord[1]-curXYCoord[1]):, :]
      maxX = min(len(previousFrame[0]), len(curFrame[0]))
      maxY = min(len(previousFrame), len(curFrame))

      previousFrame = previousFrame[:maxY, :maxX]
      curFrame      = curFrame[:maxY, :maxX]

      # Possible optimization in the future: refine the ROI based on halfDiameterRoiBoutDetect !!!

      res = cv2.absdiff(previousFrame, curFrame)
      ret, res = cv2.threshold(res,self._hyperparameters["thresForDetectMovementWithRawVideo"],255,cv2.THRESH_BINARY)

      totDiff = cv2.countNonZero(res)
      for animalId in range(self._hyperparameters["nbAnimalsPerWell"]):
        if totDiff > self._hyperparameters["minNbPixelForDetectMovementWithRawVideo"]:
          self._auDessusPerAnimalId[animalId][i-self._firstFrame] = 1
        else:
          self._auDessusPerAnimalId[animalId][i-self._firstFrame] = 0
    else:
      self._auDessusPerAnimalId[animalId][i-self._firstFrame] = 0
    previousFrames.put(initialCurFrame)
    previousXYCoords.put([xHead, yHead])

  def _loadDLModel(self):
    # Reloading DL model for tracking with DL
    from zebrazoom.code.deepLearningFunctions.loadDLmodel import loadDLmodel
    return loadDLmodel(self._hyperparameters["trackingDL"], self._hyperparameters["unet"])

  def _trackingDL(self, wellNumber, device, dlModel):
    import torch
    debugPlus = False
    # dicotomySearchOfOptimalBlobArea = self._hyperparameters["trackingDLdicotomySearchOfOptimalBlobArea"] # 500 # 700 # 850
    # applySimpleThresholdOnPredictedMask = self._hyperparameters["applySimpleThresholdOnPredictedMask"] # 230

    xtop = self._wellPositions[wellNumber]['topLeftX']
    ytop = self._wellPositions[wellNumber]['topLeftY']
    lenX = self._wellPositions[wellNumber]['lengthX']
    lenY = self._wellPositions[wellNumber]['lengthY']

    cap = zzVideoReading.VideoCapture(self._videoPath)
    if (cap.isOpened()== False):
      print("Error opening video stream or file")
    frame_width  = int(cap.get(3))
    frame_height = int(cap.get(4))

    # Performing the tracking on each frame
    applyQuantile = self._hyperparameters["applyQuantileInDLalgo"]
    i = self._firstFrame
    cap.set(1, self._firstFrame)
    if int(self._hyperparameters["onlyDoTheTrackingForThisNumberOfFrames"]) != 0:
      self._lastFrame = min(self._lastFrame, self._firstFrame + int(self._hyperparameters["onlyDoTheTrackingForThisNumberOfFrames"]))
    while (i < self._lastFrame+1):

      if (self._hyperparameters["freqAlgoPosFollow"] != 0) and (i % self._hyperparameters["freqAlgoPosFollow"] == 0):
        print("Tracking: wellNumber:",wellNumber," ; frame:",i)
        if self._hyperparameters["popUpAlgoFollow"]:
          from zebrazoom.code.popUpAlgoFollow import prepend
          prepend("Tracking: wellNumber:" + str(wellNumber) + " ; frame:" + str(i))
      if self._hyperparameters["debugTracking"]:
        print("frame:",i)

      ret, frame = cap.read()
      if applyQuantile:
        frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
        quartileChose = 0.03
        lowVal  = int(np.quantile(frame, quartileChose))
        highVal = int(np.quantile(frame, 1 - quartileChose))
        frame[frame < lowVal]  = lowVal
        frame[frame > highVal] = highVal
        frame = frame - lowVal
        mult  = np.max(frame)
        frame = frame * (255/mult)
        frame = frame.astype('uint8')

      if not(ret):
        currentFrameNum = int(cap.get(1))
        while not(ret):
          currentFrameNum = currentFrameNum - 1
          cap.set(1, currentFrameNum)
          ret, frame = cap.read()
          if applyQuantile:
            frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
            quartileChose = 0.01
            lowVal  = int(np.quantile(frame, quartileChose))
            highVal = int(np.quantile(frame, 1 - quartileChose))
            frame[frame < lowVal]  = lowVal
            frame[frame > highVal] = highVal
            frame = frame - lowVal
            mult  = np.max(frame)
            frame = frame * (255/mult)
            frame = frame.astype('uint8')

      if applyQuantile:
        grey = frame
      else:
        grey = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)

      curFrame = grey[ytop:ytop+lenY, xtop:xtop+lenX]

      if self._hyperparameters["unet"]:

        device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        imgTorch = torch.from_numpy(curFrame/255)
        imgTorch = imgTorch.unsqueeze(0).unsqueeze(0)
        imgTorch = imgTorch.to(device=device, dtype=torch.float32)

        output = dlModel(imgTorch).cpu()

        import torch.nn.functional as F
        output = F.interpolate(output, (len(curFrame[1]), len(curFrame)), mode='bilinear')
        if dlModel.n_classes > 1:
            mask = output.argmax(dim=1)
        else:
            mask = torch.sigmoid(output) > out_threshold
        thresh2 = mask[0].long().squeeze().numpy()
        thresh2 = thresh2 * 255
        thresh2 = thresh2.astype('uint8')

        thresh3 = thresh2.copy()

        if self._hyperparameters["debugTracking"]:
          self._debugFrame(thresh2, title='After Unet')

        lastFirstTheta = self._headTrackingHeadingCalculation(i, thresh2, thresh2, thresh2, thresh2, self._hyperparameters["erodeSize"], frame_width, frame_height, self._trackingHeadingAllAnimals, self._trackingHeadTailAllAnimals, self._trackingProbabilityOfGoodDetection, 0, self._wellPositions[wellNumber]["lengthX"])

        if self._hyperparameters["trackTail"] == 1:
          for animalId in range(0, self._hyperparameters["nbAnimalsPerWell"]):
            self._tailTracking(animalId, i, thresh3, thresh3, thresh3, 0, self._trackingHeadTailAllAnimals, self._trackingHeadingAllAnimals, 0, 0, 0, thresh3, 0, wellNumber)

        # Debug functions
        self._debugTracking(i, self._trackingHeadTailAllAnimals, self._trackingHeadingAllAnimals, curFrame)

      else: # mask rcnn

        oneChannel  = curFrame.tolist()
        oneChannel2 = [[a/255 for a in list] for list in oneChannel]
        imgTorch    = torch.tensor([oneChannel2, oneChannel2, oneChannel2])

        with torch.no_grad():
          prediction = dlModel([imgTorch.to(device)])

        if len(prediction) and len(prediction[0]['masks']):
          thresh = prediction[0]['masks'][0, 0].mul(255).byte().cpu().numpy()
          if debugPlus:
            self._debugFrame(255 - thresh, title="thresh")

          if self._hyperparameters["applySimpleThresholdOnPredictedMask"]:
            ret, thresh2 = cv2.threshold(thresh, self._hyperparameters["applySimpleThresholdOnPredictedMask"], 255, cv2.THRESH_BINARY)
            if self._hyperparameters["simpleThresholdCheckMinForMaxCountour"]:
              contours, hierarchy = cv2.findContours(thresh2, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
              maxContourArea = 0
              for contour in contours:
                area = cv2.contourArea(contour)
                if area > maxContourArea:
                  maxContourArea = area
              if maxContourArea < self._hyperparameters["simpleThresholdCheckMinForMaxCountour"]:
                print("maxContour found had a value that's too low (for wellNumber:", wellNumber, ", frame:", i,")")
                countNonZeroTarget = self._hyperparameters["trackingDLdicotomySearchOfOptimalBlobArea"]
                countNonZero       = 0
                low  = 0
                high = 255
                while abs(countNonZero - countNonZeroTarget) > 100 and (high - low) > 1:
                  thresValueToTry = int((low + high) / 2)
                  ret, thresh2 = cv2.threshold(thresh, thresValueToTry, 255, cv2.THRESH_BINARY)
                  countNonZero = cv2.countNonZero(thresh2)
                  if countNonZero > countNonZeroTarget:
                    low = thresValueToTry
                  else:
                    high = thresValueToTry
          else:
            if self._hyperparameters["trackingDLdicotomySearchOfOptimalBlobArea"]:
              countNonZeroTarget = self._hyperparameters["trackingDLdicotomySearchOfOptimalBlobArea"]
              countNonZero       = 0
              low  = 0
              high = 255
              while abs(countNonZero - countNonZeroTarget) > 100 and (high - low) > 1:
                thresValueToTry = int((low + high) / 2)
                ret, thresh2 = cv2.threshold(thresh, thresValueToTry, 255, cv2.THRESH_BINARY)
                countNonZero = cv2.countNonZero(thresh2)
                if countNonZero > countNonZeroTarget:
                  low = thresValueToTry
                else:
                  high = thresValueToTry
            else:
              thresh2 = thresh

          thresh3 = thresh2.copy()

          if debugPlus:
            self._debugFrame(255 - thresh2, title="thresh2")

          if self._hyperparameters["headEmbeded"] and os.path.exists(self._videoPath+'HP.csv'):
            headPositionFirstFrame = self._getHeadPositionByFileSaved()
          else:
            headPositionFirstFrame = 0
          if self._hyperparameters["headEmbeded"] and os.path.exists(self._videoPath+'.csv'):
            tailTipFirstFrame = self._getTailTipByFileSaved()
          else:
            tailTipFirstFrame = 0
          if self._hyperparameters["headEmbeded"]:
            maxDepth = math.sqrt((headPositionFirstFrame[0] - tailTipFirstFrame[0])**2 + (headPositionFirstFrame[1] - tailTipFirstFrame[0])**2)
            maxDepth = 270 # Hack: need to change this
          else:
            maxDepth = 0

          lastFirstTheta = self._headTrackingHeadingCalculation(i, thresh2, thresh2, thresh2, thresh2, self._hyperparameters["erodeSize"], frame_width, frame_height, self._trackingHeadingAllAnimals, self._trackingHeadTailAllAnimals, self._trackingProbabilityOfGoodDetection, headPositionFirstFrame, self._wellPositions[wellNumber]["lengthX"])

          if self._hyperparameters["trackTail"] == 1:
            for animalId in range(0, self._hyperparameters["nbAnimalsPerWell"]):
              self._tailTracking(animalId, i, 255 - thresh3 if self._hyperparameters["headEmbeded"] else thresh3, thresh3, thresh3, 0, self._trackingHeadTailAllAnimals, self._trackingHeadingAllAnimals, 0, maxDepth, tailTipFirstFrame, thresh3, 0, wellNumber)

          # Debug functions
          self._debugTracking(i, self._trackingHeadTailAllAnimals, self._trackingHeadingAllAnimals, curFrame)

        else:

          print("No predictions for frame", i, "and well number", wellNumber)

      print("done for frame", i)
      i = i + 1

    return [self._trackingHeadTailAllAnimals, self._trackingHeadingAllAnimals, 0, 0, 0]

  def runTracking(self, wellNumber):
    if self._background is None:
      self._background = self.getBackground()
    if self._hyperparameters["trackingDL"]:
      import torch
      return self._trackingDL(wellNumber, torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu'), self._loadDLModel())

    if self._hyperparameters["fishTailTrackingDifficultBackground"]:
      [self._trackingHeadTailAllAnimals, self._trackingHeadingAllAnimals, self._trackingEyesAllAnimals] = self._fishTailTrackingDifficultBackground(wellNumber)
      return [self._trackingHeadTailAllAnimals, self._trackingHeadingAllAnimals, self._trackingEyesAllAnimals, 0, 0]

    if (self._hyperparameters["freqAlgoPosFollow"] != 0):
      print("lastFrame:",self._lastFrame)

    thetaDiffAccept = 1.2 # 0.5 for the head embedded maybe
    maxDepth = 0

    cap = zzVideoReading.VideoCapture(self._videoPath)
    if (cap.isOpened()== False):
      print("Error opening video stream or file")

    heading = -1
    if self._hyperparameters["headEmbeded"]:
      heading = 0.7

    threshForBlackFrames = self._getThresForBlackFrame() # For headEmbededTeresaNicolson
    cap.set(1, self._firstFrame)

    # Using the first frame of the video to calculate parameters that will be used afterwards for the tracking
    if (self._hyperparameters["headEmbeded"] == 1):
      # Getting images

      if self._hyperparameters["headEmbededRemoveBack"] == 0 and self._hyperparameters["headEmbededAutoSet_BackgroundExtractionOption"] == 0:
        [frame, thresh1] = self.headEmbededFrame(self._firstFrame, wellNumber)
      else:
        self._hyperparameters["headEmbededRemoveBack"] = 1
        self._hyperparameters["minPixelDiffForBackExtract"] = self._hyperparameters["headEmbededAutoSet_BackgroundExtractionOption"]
        [frame, thresh1] = self.headEmbededFrameBackExtract(self._background, self._firstFrame, wellNumber)

      # Setting self._hyperparameters in order to add line on image
      if self._hyperparameters["addBlackLineToImg_Width"]:
        self._addBlackLineToImgSetParameters(frame)

      # (x, y) coordinates for both eyes for head embedded fish eye tracking
      if self._hyperparameters["eyeTracking"] and self._hyperparameters["headEmbeded"] == 1:
        forEye = self.getAccentuateFrameForManualPointSelect(frame)
        if True:
          leftEyeCoordinate = list(self._getCoordinates(np.uint8(forEye * 255), "Click on the center of the left eye", True, True))
          rightEyeCoordinate = list(self._getCoordinates(np.uint8(forEye * 255), "Click on the center of the right eye", True, True))
        else:
          leftEyeCoordinate  = [261, 201] # [267, 198] # [210, 105]
          rightEyeCoordinate = [285, 157] # [290, 151] # [236, 72]
        print("leftEyeCoordinate:", leftEyeCoordinate)
        print("rightEyeCoordinate:", rightEyeCoordinate)

      # if self._hyperparameters["invertBlackWhiteOnImages"]:
        # frame   = 255 - frame

      gray = frame.copy()

      oppHeading = (heading + math.pi) % (2 * math.pi)

      # Getting headPositionFirstFrame and tailTipFirstFrame positions
      if os.path.exists(self._videoPath+'HP.csv'):
        self._headPositionFirstFrame = self._getHeadPositionByFileSaved()
      else:
        if self._hyperparameters["findHeadPositionByUserInput"]:
          frameForManualPointSelection = self.getAccentuateFrameForManualPointSelect(frame)
          self._headPositionFirstFrame = self.findHeadPositionByUserInput(frameForManualPointSelection, self._firstFrame, wellNumber)
        else:
          [frame, gray, thresh1, blur, thresh2, frame2, initialCurFrame, back, xHead, yHead] = self._getImages(cap, self._firstFrame, wellNumber)
          cap.set(1, self._firstFrame)
          lastFirstTheta = self._headTrackingHeadingCalculation(self._firstFrame, blur, thresh1, thresh2, gray, self._hyperparameters["erodeSize"], int(cap.get(3)), int(cap.get(4)), self._trackingHeadingAllAnimals, self._trackingHeadTailAllAnimals, self._trackingProbabilityOfGoodDetection, self._headPositionFirstFrame, self._wellPositions[wellNumber]["lengthX"])
      if os.path.exists(self._videoPath+'.csv'):
        self._tailTipFirstFrame  = self._getTailTipByFileSaved()
      else:
        frameForManualPointSelection = self.getAccentuateFrameForManualPointSelect(frame)
        self._tailTipFirstFrame  = self.findTailTipByUserInput(frameForManualPointSelection, self._firstFrame, wellNumber)
      if self._hyperparameters["automaticallySetSomeOfTheHeadEmbededHyperparameters"] == 1:
        self._adjustHeadEmbededHyperparameters(frame)
      # Getting max depth
      if self._hyperparameters["headEmbededTeresaNicolson"] == 1:
        if len(self._headPositionFirstFrame) == 0:
          self._headPositionFirstFrame = [self._trackingHeadTailAllAnimals[0][0][0][0], self._trackingHeadTailAllAnimals[0][0][0][1]]
        maxDepth = self._headEmbededTailTrackFindMaxDepthTeresaNicolson(frame)
      else:
        if self._hyperparameters["centerOfMassTailTracking"] == 0:
          maxDepth = self._headEmbededTailTrackFindMaxDepth(frame)
        else:
          maxDepth = self._centerOfMassTailTrackFindMaxDepth(frame)

    widgets = None
    # Performing the tracking on each frame
    i = self._firstFrame
    if int(self._hyperparameters["onlyDoTheTrackingForThisNumberOfFrames"]) != 0:
      self._lastFrame = min(self._lastFrame, self._firstFrame + int(self._hyperparameters["onlyDoTheTrackingForThisNumberOfFrames"]))
    while (i < self._lastFrame+1):

      if (self._hyperparameters["freqAlgoPosFollow"] != 0) and (i % self._hyperparameters["freqAlgoPosFollow"] == 0):
        print("Tracking: wellNumber:",wellNumber," ; frame:",i)
        if self._hyperparameters["popUpAlgoFollow"]:
          from zebrazoom.code.popUpAlgoFollow import prepend
          prepend("Tracking: wellNumber:" + str(wellNumber) + " ; frame:" + str(i))
      if self._hyperparameters["debugTracking"]:
        print("frame:",i)
      # Get images for frame i
      [frame, gray, thresh1, blur, thresh2, frame2, initialCurFrame, back, xHead, yHead] = self._getImages(cap, i, wellNumber, 0, self._trackingHeadTailAllAnimals)
      # Head tracking and heading calculation
      lastFirstTheta = self._headTrackingHeadingCalculation(i, blur, thresh1, thresh2, gray, self._hyperparameters["erodeSize"], int(cap.get(3)), int(cap.get(4)), self._trackingHeadingAllAnimals, self._trackingHeadTailAllAnimals, self._trackingProbabilityOfGoodDetection, self._headPositionFirstFrame, self._wellPositions[wellNumber]["lengthX"], xHead, yHead)

      # Tail tracking for frame i
      if self._hyperparameters["trackTail"] == 1 :
        for animalId in range(0, self._hyperparameters["nbAnimalsPerWell"]):
          self._tailTracking(animalId, i, frame, thresh1, threshForBlackFrames, thetaDiffAccept, self._trackingHeadTailAllAnimals, self._trackingHeadingAllAnimals, lastFirstTheta, maxDepth, self._tailTipFirstFrame, initialCurFrame, back, wellNumber, xHead, yHead)

      if self._hyperparameters["updateBackgroundAtInterval"]:
        self._updateBackgroundAtInterval(i, wellNumber, initialCurFrame, self._trackingHeadTailAllAnimals, frame)

      # Eye tracking for frame i
      if self._hyperparameters["eyeTracking"]:
        if self._hyperparameters["headEmbeded"] == 1:
          if self._hyperparameters["adjustHeadEmbeddedEyeTracking"]:
            i, widgets = self._eyeTrackingHeadEmbedded(animalId, i, frame, thresh1, self._trackingHeadingAllAnimals, self._trackingHeadTailAllAnimals, self._trackingEyesAllAnimals, leftEyeCoordinate, rightEyeCoordinate, widgets=widgets)
            if not self._hyperparameters["eyeFilterKernelSize"] % 2:
              self._hyperparameters["eyeFilterKernelSize"] -= 1
            continue
          else:
            self._eyeTrackingHeadEmbedded(animalId, i, frame, thresh1, self._trackingHeadingAllAnimals, self._trackingHeadTailAllAnimals, self._trackingEyesAllAnimals, leftEyeCoordinate, rightEyeCoordinate)
        else:
          self._eyeTracking(animalId, i, frame, thresh1, self._trackingHeadingAllAnimals, self._trackingHeadTailAllAnimals, self._trackingEyesAllAnimals)

      # Debug functions
      if self._hyperparameters["nbAnimalsPerWell"] > 1 or self._hyperparameters["forceBlobMethodForHeadTracking"] or self._hyperparameters["headEmbeded"] == 1 or self._hyperparameters["fixedHeadPositionX"] != -1:
        self._debugTracking(i, self._trackingHeadTailAllAnimals, self._trackingHeadingAllAnimals, frame)
      else:
        self._debugTracking(i, self._trackingHeadTailAllAnimals, self._trackingHeadingAllAnimals, frame)
      # DetectMovementWithRawVideoInsideTracking
      if self._hyperparameters["detectMovementWithRawVideoInsideTracking"]:
        self._detectMovementWithRawVideoInsideTracking(i, xHead, yHead, initialCurFrame)

      if self._hyperparameters["trackOnlyOnROI_halfDiameter"]:
        if not(xHead == 0 and yHead == 0):
          for animalId in range(0, self._hyperparameters["nbAnimalsPerWell"]):
            for j in range(0, len(self._trackingHeadTailAllAnimals[animalId][i-self._firstFrame])): # Head Position should already shifted, only shifting tail positions now
              self._trackingHeadTailAllAnimals[animalId][i-self._firstFrame][j][0] = self._trackingHeadTailAllAnimals[animalId][i-self._firstFrame][j][0] + xHead
              self._trackingHeadTailAllAnimals[animalId][i-self._firstFrame][j][1] = self._trackingHeadTailAllAnimals[animalId][i-self._firstFrame][j][1] + yHead

      paramsAdjusted = self._adjustParameters(i, initialCurFrame, frame, frame2, widgets)
      if paramsAdjusted is not None:
        i, widgets = paramsAdjusted
      else:
        i = i + 1

    if self._hyperparameters["postProcessMultipleTrajectories"]:
      self._postProcessMultipleTrajectories(self._trackingHeadTailAllAnimals, self._trackingProbabilityOfGoodDetection)

    self._savingBlackFrames(self._trackingHeadTailAllAnimals)

    print("Tracking done for well", wellNumber)
    if self._hyperparameters["popUpAlgoFollow"]:
      from zebrazoom.code.popUpAlgoFollow import prepend
      prepend("Tracking done for well "+ str(wellNumber))

    if self._auDessusPerAnimalId is not None:
      return [self._trackingHeadTailAllAnimals, self._trackingHeadingAllAnimals, self._trackingEyesAllAnimals, self._headPositionFirstFrame, self._tailTipFirstFrame, self._auDessusPerAnimalId]
    else:
      return [self._trackingHeadTailAllAnimals, self._trackingHeadingAllAnimals, self._trackingEyesAllAnimals, self._headPositionFirstFrame, self._tailTipFirstFrame]

  def _reloadPreviousTrackingData(self):
    # Reloading previously extracted tracking data if debugging option selected
    with open(os.path.join(self._hyperparameters["outputFolder"], self._videoName, 'intermediaryTracking.txt'), 'rb') as outfile:
        return pickle.load(outfile)

  def _getParametersForWell(self, wellNumber):
    '''Does the tracking and then the extraction of parameters'''
    if self.useGUI:
      from PyQt5.QtWidgets import QApplication

      if QApplication.instance() is None:
        from zebrazoom.GUIAllPy import PlainApplication
        app = PlainApplication(sys.argv)
    if self._hyperparameters["debugPauseBetweenTrackAndParamExtract"] == "noDebug":
      # Normal execution process
      trackingData = self.runTracking(wellNumber)
      parameters = extractParameters(trackingData, wellNumber, self._hyperparameters, self._videoPath, self._wellPositions, self._background)
      return [wellNumber,parameters,[]]
    elif self._hyperparameters["debugPauseBetweenTrackAndParamExtract"] == "justSaveTrackData":
      # Extracing tracking data, saving it, and that's it
      trackingData = self.runTracking(wellNumber)
      return [wellNumber,[],trackingData]
    elif self._hyperparameters["debugPauseBetweenTrackAndParamExtract"] == "saveTrackDataAndExtractParam":
      # Extracing tracking data, saving it, and continuing normal execution
      trackingData = self.runTracking(wellNumber)
      parameters = extractParameters(trackingData, wellNumber, self._hyperparameters, self._videoPath, self._wellPositions, self._background)
      return [wellNumber,parameters,trackingData]
    else:
      assert self._hyperparameters["debugPauseBetweenTrackAndParamExtract"] == "justExtractParamFromPreviousTrackData"
      # Reloading previously tracked data and using it to extract parameters
      trackingData = self._reloadPreviousTrackingData()[wellNumber]
      parameters = extractParameters(trackingData, wellNumber, self._hyperparameters, self._videoPath, self._wellPositions, self._background)
      return [wellNumber,parameters,[]]

  def _storeParametersInQueue(self, queue, wellNumber):
    queue.put(self._getParametersForWell(wellNumber))

  def _storeIntermediaryResults(self, trackingDataPerWell):
    # saving tracking results for future uses
    with open(os.path.join(self._hyperparameters["outputFolder"], self._videoName, 'intermediaryTracking.txt'),'wb') as outfile:
      pickle.dump(trackingDataPerWell, outfile)
    if (self._hyperparameters["freqAlgoPosFollow"] != 0):
      print("intermediary results saved")

  def run(self):
    self._background = self._getBackground()

    if self._hyperparameters["trackingDL"]:
      from torch.multiprocessing import Process
      import torch.multiprocessing as mp
    else:
      from multiprocessing import Process
      import multiprocessing as mp
    if globalVariables["mac"] or self._hyperparameters["trackingDL"]:
      mp.set_start_method('spawn', force=True)

    # Tracking and extraction of parameters
    if globalVariables["noMultiprocessing"] == 0 and not self._hyperparameters['headEmbeded']:
      if self._hyperparameters["onlyTrackThisOneWell"] == -1:
        # for all wells, in parallel
        queue = mp.Queue()
        processes = [Process(target=self._storeParametersInQueue, args=(queue, wellNumber), daemon=True)
                     for wellNumber in range(self._hyperparameters["nbWells"])]
        for p in processes:
          p.start()
        parametersPerWell = [queue.get() for p in processes]
        for p in processes:
          p.join()
      else:
        # for just one well
        parametersPerWell = [self._getParametersForWell(self._hyperparameters["onlyTrackThisOneWell"])]
    else:
      if self._hyperparameters["onlyTrackThisOneWell"] == -1:
        parametersPerWell = [self._getParametersForWell(wellNumber) for wellNumber in range(self._hyperparameters["nbWells"])]
      else:
        parametersPerWell = [self._getParametersForWell(self._hyperparameters["onlyTrackThisOneWell"])]

    if (self._hyperparameters["debugPauseBetweenTrackAndParamExtract"] == "saveTrackDataAndExtractParam") or (self._hyperparameters["debugPauseBetweenTrackAndParamExtract"] == "justSaveTrackData"):
      self._storeIntermediaryResults({wellNumber: trackingData for wellNumber, _, trackingData in parametersPerWell})

    # Sorting wells after the end of the parallelized calls end
    return {wellNumber: parameters for wellNumber, parameters, _ in parametersPerWell}

register_tracking_method('tracking', Tracking)



import math

import cv2
import numpy as np

from ._computeHeading import ComputeHeadingMixin
from ._headTrackingHeadingCalculation import HeadTrackingHeadingCalculationMixin


class EyeTrackingMixin(ComputeHeadingMixin, HeadTrackingHeadingCalculationMixin):
  def _adjustHeadEmbeddedEyeTrackingParamsSegment(self, i, colorFrame, widgets):
    pass

  def _adjustHeadEmbeddedEyeTrackingParamsSegment(self, i, colorFrame, widgets):
    pass

  def _eyeTrackingHeadEmbedded(self, animalId, i, frame, thresh1, trackingHeadingAllAnimals, trackingHeadTailAllAnimals, trackingEyesAllAnimals, leftEyeCoordinate, rightEyeCoordinate, widgets=None):
    if self._hyperparameters["eyeTrackingHeadEmbeddedWithSegment"]:
      return self._eyeTrackingHeadEmbeddedSegment(animalId, i, frame, thresh1, trackingHeadingAllAnimals, trackingHeadTailAllAnimals, trackingEyesAllAnimals, leftEyeCoordinate, rightEyeCoordinate, widgets)

    if self._hyperparameters["eyeTrackingHeadEmbeddedWithEllipse"]:
      return self._eyeTrackingHeadEmbeddedEllipse(animalId, i, frame, thresh1, trackingHeadingAllAnimals, trackingHeadTailAllAnimals, trackingEyesAllAnimals, leftEyeCoordinate, rightEyeCoordinate, widgets)

  def _eyeTrackingHeadEmbeddedSegment(self, animalId, i, frame, thresh1, trackingHeadingAllAnimals, trackingHeadTailAllAnimals, trackingEyesAllAnimals, leftEyeCoordinate, rightEyeCoordinate, widgets):
    headingLineHalfDiameter = self._hyperparameters["eyeTrackingHeadEmbeddedHalfDiameter"]
    headingLineWidthLeft    = self._hyperparameters["eyeTrackingHeadEmbeddedWidthLeft"] if self._hyperparameters["eyeTrackingHeadEmbeddedWidthLeft"] else self._hyperparameters["eyeTrackingHeadEmbeddedWidth"]
    headingLineWidthRight   = self._hyperparameters["eyeTrackingHeadEmbeddedWidthRight"] if self._hyperparameters["eyeTrackingHeadEmbeddedWidthRight"] else self._hyperparameters["eyeTrackingHeadEmbeddedWidth"]
    headingLineWidthArray = [headingLineWidthLeft, headingLineWidthRight]

    if self._hyperparameters["improveContrastForEyeDetectionOfHeadEmbedded"]:
      forEye = self.getAccentuateFrameForManualPointSelect(frame) * 255
    else:
      forEye = frame.copy()
    if self._hyperparameters["invertColorsForHeadEmbeddedEyeTracking"]:
      forEye = 255 - forEye

    forEye = forEye.astype(np.uint8)

    angle = []
    for eyeIdx, eyeCoordinate in enumerate([leftEyeCoordinate, rightEyeCoordinate]):
      headingLineWidth = headingLineWidthArray[eyeIdx]
      forSpecificEye = forEye[int(eyeCoordinate[1]-headingLineHalfDiameter):int(eyeCoordinate[1]+headingLineHalfDiameter), int(eyeCoordinate[0]-headingLineHalfDiameter):int(eyeCoordinate[0]+headingLineHalfDiameter)]
      pixelSum  = 0
      bestAngle = 0
      nTries    = 20
      for j in range(0, nTries):
        angleOption = j * (math.pi / nTries)
        startPoint = (int(headingLineHalfDiameter - headingLineHalfDiameter * math.cos(angleOption)), int(headingLineHalfDiameter - headingLineHalfDiameter * math.sin(angleOption)))
        endPoint   = (int(headingLineHalfDiameter + headingLineHalfDiameter * math.cos(angleOption)), int(headingLineHalfDiameter + headingLineHalfDiameter * math.sin(angleOption)))
        testImage  = forSpecificEye.copy()
        testImage  = cv2.line(testImage, startPoint, endPoint, (0), headingLineWidth)
        nbWhitePixels = np.sum(testImage)
        if nbWhitePixels > pixelSum:
          pixelSum  = nbWhitePixels
          bestAngle = angleOption
      bestAngle1 = bestAngle
      nTries2     = 50
      for j2 in range(0, nTries2):
        angleOption = bestAngle1 - ((math.pi / nTries) / 2) + ((j2 / nTries2) * (math.pi / nTries))
        startPoint = (int(headingLineHalfDiameter - headingLineHalfDiameter * math.cos(angleOption)), int(headingLineHalfDiameter - headingLineHalfDiameter * math.sin(angleOption)))
        endPoint   = (int(headingLineHalfDiameter + headingLineHalfDiameter * math.cos(angleOption)), int(headingLineHalfDiameter + headingLineHalfDiameter * math.sin(angleOption)))
        testImage  = forSpecificEye.copy()
        testImage  = cv2.line(testImage, startPoint, endPoint, (0), headingLineWidth)
        nbWhitePixels = np.sum(testImage)
        if nbWhitePixels > pixelSum:
          pixelSum  = nbWhitePixels
          bestAngle = angleOption
      angle.append(bestAngle)

    leftEyeAngle  = angle[0]
    rightEyeAngle = angle[1]

    # Debugging Plot
    if self._hyperparameters["debugEyeTracking"] or self._hyperparameters["adjustHeadEmbeddedEyeTracking"]:
      if self._hyperparameters["improveContrastForEyeDetectionOfHeadEmbedded"]:
        forEye2 = self.getAccentuateFrameForManualPointSelect(frame) * 255
      else:
        forEye2 = frame.copy()
      if self._hyperparameters["invertColorsForHeadEmbeddedEyeTracking"]:
        forEye2 = 255 - forEye2
      forEye2 = forEye2.astype(np.uint8)
      colorFrame = forEye2.copy()
      # Left eye
      cv2.circle(colorFrame, (leftEyeCoordinate[0], leftEyeCoordinate[1]), 2, (0,255,255), 1)
      cv2.line(colorFrame, (int(leftEyeCoordinate[0]-headingLineHalfDiameter*math.cos(leftEyeAngle)), int(leftEyeCoordinate[1]-headingLineHalfDiameter*math.sin(leftEyeAngle))), (int(leftEyeCoordinate[0]+headingLineHalfDiameter*math.cos(leftEyeAngle)), int(leftEyeCoordinate[1]+headingLineHalfDiameter*math.sin(leftEyeAngle))), (255,0,255), headingLineWidthLeft)
      # Right eye
      cv2.circle(colorFrame, (rightEyeCoordinate[0], rightEyeCoordinate[1]), 2, (0,255,255), 1)
      cv2.line(colorFrame, (int(rightEyeCoordinate[0]-headingLineHalfDiameter*math.cos(rightEyeAngle)), int(rightEyeCoordinate[1]-headingLineHalfDiameter*math.sin(rightEyeAngle))), (int(rightEyeCoordinate[0]+headingLineHalfDiameter*math.cos(rightEyeAngle)), int(rightEyeCoordinate[1]+headingLineHalfDiameter*math.sin(rightEyeAngle))), (255,0,255), headingLineWidthRight)
      if self._hyperparameters["debugEyeTracking"]:
        self._debugFrame(colorFrame, title='Eye Tracking debugging')
      if self._hyperparameters["adjustHeadEmbeddedEyeTracking"]:
        return self._adjustHeadEmbeddedEyeTrackingParamsSegment(i, colorFrame, widgets)

    trackingEyesAllAnimals[animalId, i-self._firstFrame, 0] = leftEyeCoordinate[0]
    trackingEyesAllAnimals[animalId, i-self._firstFrame, 1] = leftEyeCoordinate[1]
    trackingEyesAllAnimals[animalId, i-self._firstFrame, 2] = leftEyeAngle
    trackingEyesAllAnimals[animalId, i-self._firstFrame, 3] = 0
    trackingEyesAllAnimals[animalId, i-self._firstFrame, 4] = rightEyeCoordinate[0]
    trackingEyesAllAnimals[animalId, i-self._firstFrame, 5] = rightEyeCoordinate[1]
    trackingEyesAllAnimals[animalId, i-self._firstFrame, 6] = rightEyeAngle
    trackingEyesAllAnimals[animalId, i-self._firstFrame, 7] = 0

  def _eyeTrackingHeadEmbeddedEllipse(self, animalId, i, frame, thresh1, trackingHeadingAllAnimals, trackingHeadTailAllAnimals, trackingEyesAllAnimals, leftEyeCoordinate, rightEyeCoordinate, widgets):
    if self._hyperparameters["improveContrastForEyeDetectionOfHeadEmbedded"]:
      forEye = self.getAccentuateFrameForManualPointSelect(frame) * 255
    else:
      forEye = frame.copy()
    if self._hyperparameters["invertColorsForHeadEmbeddedEyeTracking"]:
      forEye = 255 - forEye

    forEye = forEye.astype(np.uint8)

    eyeBinaryThreshold = self._hyperparameters["eyeBinaryThreshold"] # 15

    ret, threshEye = cv2.threshold(forEye, eyeBinaryThreshold, 255, cv2.THRESH_BINARY)

    kernel = np.ones((self._hyperparameters["eyeFilterKernelSize"], self._hyperparameters["eyeFilterKernelSize"]), np.uint8)
    threshEye = cv2.erode(threshEye, kernel, iterations=1)
    threshEye = cv2.dilate(threshEye, kernel, iterations=1)

    threshEye[0,:] = 255
    threshEye[len(threshEye)-1,:] = 255
    threshEye[:,0] = 255
    threshEye[:,len(threshEye[0])-1] = 255

    contourLeft  = 0
    contourRight = 0

    contours, hierarchy = cv2.findContours(threshEye, cv2.RETR_LIST, cv2.CHAIN_APPROX_SIMPLE)
    maxLeft  = 1000000000000000000000000000
    maxRight = 1000000000000000000000000000
    maxArea  = len(threshEye) * len(threshEye[0]) * 0.9
    for contour in contours:
      area = cv2.contourArea(contour)
      M = cv2.moments(contour)
      if M['m00']:
        cx = int(M['m10']/M['m00'])
        cy = int(M['m01']/M['m00'])
      else:
        cx = 0
        cy = 0
      if area < maxArea:
        if (leftEyeCoordinate[0] - cx)**2 + (leftEyeCoordinate[1] - cy)**2 < maxLeft:
          maxLeft  = (leftEyeCoordinate[0] - cx)**2 + (leftEyeCoordinate[1] - cy)**2
          contourLeft  = contour
        if (rightEyeCoordinate[0] - cx)**2 + (rightEyeCoordinate[1] - cy)**2 < maxRight:
          maxRight = (rightEyeCoordinate[0] - cx)**2 + (rightEyeCoordinate[1] - cy)**2
          contourRight = contour

    # Finding the (X, Y) coordinates and the angle of each of the two eyes
    eyeAngle = [0, 0]
    eyeX = [leftEyeCoordinate[0], rightEyeCoordinate[0]]
    eyeY = [leftEyeCoordinate[1], rightEyeCoordinate[1]]
    for idx, contour in enumerate([contourLeft, contourRight]):
      if type(contour) != int and len(contour) >= 3:
        if len(contour) >= 5:
          ellipse = cv2.fitEllipse(contour)
          angle1 = ellipse[2] * (math.pi / 180) + (math.pi / 2)
          if self._hyperparameters["debugEyeTrackingAdvanced"]:
            threshEye1 = np.zeros((len(threshEye), len(threshEye[0])))
            threshEye1[:, :] = 0
            threshEye1 = threshEye1.astype(np.uint8)
            cv2.fillPoly(threshEye1, pts =[contour], color=(255))
        else:
          print("problem with eye angle here, not enough points in the contour to use fitEllipse")
          threshEye1 = np.zeros((len(threshEye), len(threshEye[0])))
          threshEye1[:, :] = 0
          cv2.fillPoly(threshEye1, pts =[contour], color=(255))
          if self._hyperparameters["debugEyeTrackingAdvanced"]:
            self._debugFrame(threshEye1, title='Frame')

          minWhitePixel = 10000000000000000000000
          bestAngle     = 0
          nTries        = 200
          for j in range(0, nTries):
            angleOption = j * (math.pi / nTries)
            startPoint = (int(eyeX[idx] - 100000 * math.cos(angleOption)), int(eyeY[idx] - 100000 * math.sin(angleOption)))
            endPoint   = (int(eyeX[idx] + 100000 * math.cos(angleOption)), int(eyeY[idx] + 100000 * math.sin(angleOption)))
            testImage  = threshEye1.copy()
            testImage  = cv2.line(testImage, startPoint, endPoint, (0), 4)
            nbWhitePixels = cv2.countNonZero(testImage)
            if nbWhitePixels < minWhitePixel:
              minWhitePixel = nbWhitePixels
              bestAngle     = angleOption
            if False:
              firstBestAngle = bestAngle
              for j in range(0, nTries):
                angleOption = firstBestAngle - (math.pi / 5) + j * ((2 * (math.pi / 5)) / nTries)
                startPoint = (int(eyeX[idx] - 100000 * math.cos(angleOption)), int(eyeY[idx] - 100000 * math.sin(angleOption)))
                endPoint   = (int(eyeX[idx] + 100000 * math.cos(angleOption)), int(eyeY[idx] + 100000 * math.sin(angleOption)))
                testImage  = threshEye1.copy()
                testImage  = cv2.line(testImage, startPoint, endPoint, (0), 4)
                nbWhitePixels = cv2.countNonZero(testImage)
                if nbWhitePixels < minWhitePixel:
                  minWhitePixel = nbWhitePixels
                  bestAngle     = angleOption
          angle1 = bestAngle

        headingApproximate = trackingHeadingAllAnimals[animalId, i-self._firstFrame] % (2*math.pi)
        headingPreciseOpt1 = angle1
        headingPreciseOpt2 = (angle1 + math.pi) % (2*math.pi)
        diffAngle1 = self._distBetweenThetas(headingApproximate, headingPreciseOpt1)
        diffAngle2 = self._distBetweenThetas(headingApproximate, headingPreciseOpt2)
        if (diffAngle1 < diffAngle2):
          eyeAngle[idx] = headingPreciseOpt1
        else:
          eyeAngle[idx] = headingPreciseOpt2
      else:
        eyeAngle[idx] = 0
    # Debugging Plot
    headingLineValidationPlotLength = self._hyperparameters["eyeTrackingHeadEmbeddedHalfDiameter"]
    if self._hyperparameters["debugEyeTracking"] or self._hyperparameters["adjustHeadEmbeddedEyeTracking"]:
      colorFrame = cv2.cvtColor(threshEye, cv2.COLOR_GRAY2RGB)
      cv2.circle(colorFrame, (eyeX[0], eyeY[0]), 2, (0,255,255), 1)
      cv2.line(colorFrame, (eyeX[0], eyeY[0]), (int(eyeX[0]+headingLineValidationPlotLength*math.cos(eyeAngle[0])), int(eyeY[0]+headingLineValidationPlotLength*math.sin(eyeAngle[0]))), (255,0,255), 1)
      cv2.line(colorFrame, (eyeX[1], eyeY[1]), (int(eyeX[1]+headingLineValidationPlotLength*math.cos(eyeAngle[1])), int(eyeY[1]+headingLineValidationPlotLength*math.sin(eyeAngle[1]))), (255,0,255), 1)
      cv2.circle(colorFrame, (eyeX[1], eyeY[1]), 2, (0,255,255), 1)
      if self._hyperparameters["adjustHeadEmbeddedEyeTracking"]:
        return self._adjustHeadEmbeddedEyeTrackingParamsEllipse(i, colorFrame, widgets)
      else:
        self._debugFrame(colorFrame, title='Eye Tracking debugging')

    # Storing the (X, Y) coordinates and angles
    trackingEyesAllAnimals[animalId, i-self._firstFrame, 0] = leftEyeCoordinate[0]
    trackingEyesAllAnimals[animalId, i-self._firstFrame, 1] = leftEyeCoordinate[1]
    trackingEyesAllAnimals[animalId, i-self._firstFrame, 2] = eyeAngle[0]
    trackingEyesAllAnimals[animalId, i-self._firstFrame, 3] = 0
    trackingEyesAllAnimals[animalId, i-self._firstFrame, 4] = rightEyeCoordinate[0]
    trackingEyesAllAnimals[animalId, i-self._firstFrame, 5] = rightEyeCoordinate[1]
    trackingEyesAllAnimals[animalId, i-self._firstFrame, 6] = eyeAngle[1]
    trackingEyesAllAnimals[animalId, i-self._firstFrame, 7] = 0

  def _eyeTracking(self, animalId, i, frame, thresh1, trackingHeadingAllAnimals, trackingHeadTailAllAnimals, trackingEyesAllAnimals):
    headCenterToMidEyesPointDistance   = self._hyperparameters["headCenterToMidEyesPointDistance"]
    eyeBinaryThreshold                 = self._hyperparameters["eyeBinaryThreshold"]
    midEyesPointToEyeCenterMaxDistance = self._hyperparameters["midEyesPointToEyeCenterMaxDistance"]
    eyeHeadingSearchAreaHalfDiameter   = self._hyperparameters["eyeHeadingSearchAreaHalfDiameter"]
    headingLineValidationPlotLength    = self._hyperparameters["headingLineValidationPlotLength"]
    debugEyeTracking                   = self._hyperparameters["debugEyeTracking"]
    debugEyeTrackingAdvanced           = self._hyperparameters["debugEyeTrackingAdvanced"]

    # Retrieving the X, Y coordinates of the center of the head of the fish and calculating the "mid eyes" point
    x = trackingHeadTailAllAnimals[animalId, i-self._firstFrame][0][0]
    y = trackingHeadTailAllAnimals[animalId, i-self._firstFrame][0][1]
    midEyesPointX = int(x+headCenterToMidEyesPointDistance*math.cos(trackingHeadingAllAnimals[animalId, i-self._firstFrame]))
    midEyesPointY = int(y+headCenterToMidEyesPointDistance*math.sin(trackingHeadingAllAnimals[animalId, i-self._firstFrame]))

    # Finding the connected components associated with each of the two eyes
    ret, threshEye = cv2.threshold(frame, eyeBinaryThreshold, 255, cv2.THRESH_BINARY)
    threshEye[0,:] = 255
    threshEye[len(threshEye)-1,:] = 255
    threshEye[:,0] = 255
    threshEye[:,len(threshEye[0])-1] = 255
    # Adding a white circle on the swim bladder
    whiteCircleDiameter = int(1.2 * headCenterToMidEyesPointDistance)
    whiteCircleX = int(x-whiteCircleDiameter*math.cos(trackingHeadingAllAnimals[animalId, i-self._firstFrame]))
    whiteCircleY = int(y-whiteCircleDiameter*math.sin(trackingHeadingAllAnimals[animalId, i-self._firstFrame]))
    cv2.circle(threshEye, (whiteCircleX, whiteCircleY), whiteCircleDiameter, (255, 255, 255), -1)
    maxArea1    = 0 # Biggest of the two contours
    maxContour1 = 0
    maxArea2    = 0 # Smallest of the two contours
    maxContour2 = 0
    contours, hierarchy = cv2.findContours(threshEye, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
    for contour in contours:
      area = cv2.contourArea(contour)
      M = cv2.moments(contour)
      if M['m00']:
        cx = int(M['m10']/M['m00'])
        cy = int(M['m01']/M['m00'])
      else:
        cx = 0
        cy = 0
      if area < (len(frame)*len(frame[0]))/2 and math.sqrt((cx - midEyesPointX)**2 + (cy - midEyesPointY)**2) < midEyesPointToEyeCenterMaxDistance:
        if (area > maxArea1):
          maxArea2    = maxArea1
          maxContour2 = maxContour1
          maxArea1    = area
          maxContour1 = contour
        else:
          if (area > maxArea2):
            maxArea2    = area
            maxContour2 = contour
    if debugEyeTrackingAdvanced:
      self._debugFrame(threshEye, title='Frame')
    # Finding, in the image without the white circle, the contours corresponding to the contours previously found in the image with the white circle
    # This to make sure that we get the blobs that "really" correspond to the eyes in the unlikely event where the white circle would have overlapped with the eyes
    M = cv2.moments(maxContour1)
    if M['m00']:
      eye1X = int(M['m10']/M['m00'])
      eye1Y = int(M['m01']/M['m00'])
    else:
      eye1X = 0
      eye1Y = 0
    M = cv2.moments(maxContour2)
    if M['m00']:
      eye2X = int(M['m10']/M['m00'])
      eye2Y = int(M['m01']/M['m00'])
    else:
      eye2X = 0
      eye2Y = 0
    maxContour1b = maxContour1
    maxContour2b = maxContour2
    ret, threshEye2 = cv2.threshold(frame, eyeBinaryThreshold, 255, cv2.THRESH_BINARY)
    threshEye2[0,:] = 255
    threshEye2[len(threshEye2)-1,:] = 255
    threshEye2[:,0] = 255
    threshEye2[:,len(threshEye2[0])-1] = 255
    contours, hierarchy = cv2.findContours(threshEye2, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
    for contour in contours:
      dist1 = cv2.pointPolygonTest(contour, (eye1X, eye1Y), True)
      dist2 = cv2.pointPolygonTest(contour, (eye2X, eye2Y), True)
      if dist1 >= 0:
        maxContour1b = contour
      if dist2 >= 0:
        maxContour2b = contour
    # Finding the left and the right eyes
    M = cv2.moments(maxContour1b)
    if M['m00']:
      eye1X = int(M['m10']/M['m00'])
      eye1Y = int(M['m01']/M['m00'])
    else:
      eye1X = 0
      eye1Y = 0
    M = cv2.moments(maxContour2b)
    if M['m00']:
      eye2X = int(M['m10']/M['m00'])
      eye2Y = int(M['m01']/M['m00'])
    else:
      eye2X = 0
      eye2Y = 0
    distBetweenTheTwoEyes = math.sqrt((eye1X - eye2X)**2 + (eye1Y - eye2Y)**2)
    estimatedLeftEyeX = int(midEyesPointX + (distBetweenTheTwoEyes/2) * math.cos(trackingHeadingAllAnimals[animalId, i-self._firstFrame] - (math.pi/2)))
    estimatedLeftEyeY = int(midEyesPointY + (distBetweenTheTwoEyes/2) * math.sin(trackingHeadingAllAnimals[animalId, i-self._firstFrame] - (math.pi/2)))
    contour1ToEstimatedLeftEyeYdistance = math.sqrt((estimatedLeftEyeX - eye1X)**2 + (estimatedLeftEyeY - eye1Y)**2)
    contour2ToEstimatedLeftEyeYdistance = math.sqrt((estimatedLeftEyeX - eye2X)**2 + (estimatedLeftEyeY - eye2Y)**2)
    if contour1ToEstimatedLeftEyeYdistance < contour2ToEstimatedLeftEyeYdistance:
      contourLeft  = maxContour1b
      contourRight = maxContour2b
    else:
      contourLeft  = maxContour2b
      contourRight = maxContour1b
    # Finding the (X, Y) coordinates and the angle of each of the two eyes
    eyeX     = [0, 0]
    eyeY     = [0, 0]
    eyeAngle = [0, 0]
    eyeArea  = [0, 0]
    for idx, contour in enumerate([contourLeft, contourRight]):
      eyeArea[idx] = cv2.contourArea(contour)
      M = cv2.moments(contour)
      if M['m00']:
        eyeX[idx] = int(M['m10']/M['m00'])
        eyeY[idx] = int(M['m01']/M['m00'])
      if type(contour) != int and len(contour) >= 3:
        if len(contour) >= 5:
          ellipse = cv2.fitEllipse(contour)
          angle1 = ellipse[2] * (math.pi / 180) + (math.pi / 2)
        else:
          print("problem with eye angle here, not enough points in the contour to use fitEllipse")
          threshEye1 = np.zeros((len(threshEye), len(threshEye[0])))
          threshEye1[:, :] = 0
          cv2.fillPoly(threshEye1, pts =[contour], color=(255))
          if debugEyeTrackingAdvanced:
            self._debugFrame(threshEye1, title='Frame')
          if False:
            angle1 = self._computeHeading(threshEye1, eyeX[idx], eyeY[idx], eyeHeadingSearchAreaHalfDiameter)
          else:
            minWhitePixel = 10000000000000000000000
            bestAngle     = 0
            nTries        = 200
            for j in range(0, nTries):
              angleOption = j * (math.pi / nTries)
              startPoint = (int(eyeX[idx] - 100000 * math.cos(angleOption)), int(eyeY[idx] - 100000 * math.sin(angleOption)))
              endPoint   = (int(eyeX[idx] + 100000 * math.cos(angleOption)), int(eyeY[idx] + 100000 * math.sin(angleOption)))
              testImage  = threshEye1.copy()
              testImage  = cv2.line(testImage, startPoint, endPoint, (0), 4)
              nbWhitePixels = cv2.countNonZero(testImage)
              if nbWhitePixels < minWhitePixel:
                minWhitePixel = nbWhitePixels
                bestAngle     = angleOption
              if False:
                firstBestAngle = bestAngle
                for j in range(0, nTries):
                  angleOption = firstBestAngle - (math.pi / 5) + j * ((2 * (math.pi / 5)) / nTries)
                  startPoint = (int(eyeX[idx] - 100000 * math.cos(angleOption)), int(eyeY[idx] - 100000 * math.sin(angleOption)))
                  endPoint   = (int(eyeX[idx] + 100000 * math.cos(angleOption)), int(eyeY[idx] + 100000 * math.sin(angleOption)))
                  testImage  = threshEye1.copy()
                  testImage  = cv2.line(testImage, startPoint, endPoint, (0), 4)
                  nbWhitePixels = cv2.countNonZero(testImage)
                  if nbWhitePixels < minWhitePixel:
                    minWhitePixel = nbWhitePixels
                    bestAngle     = angleOption
            angle1 = bestAngle

        headingApproximate = trackingHeadingAllAnimals[animalId, i-self._firstFrame] % (2*math.pi)
        headingPreciseOpt1 = angle1
        headingPreciseOpt2 = (angle1 + math.pi) % (2*math.pi)
        diffAngle1 = self._distBetweenThetas(headingApproximate, headingPreciseOpt1)
        diffAngle2 = self._distBetweenThetas(headingApproximate, headingPreciseOpt2)
        if (diffAngle1 < diffAngle2):
          eyeAngle[idx] = headingPreciseOpt1
        else:
          eyeAngle[idx] = headingPreciseOpt2
      else:
        eyeAngle[idx] = 0
    # Debugging Plot
    if debugEyeTracking:
      colorFrame = cv2.cvtColor(frame, cv2.COLOR_GRAY2RGB)
      cv2.line(colorFrame, (int(x),int(y)), (int(x+2*headingLineValidationPlotLength*math.cos(trackingHeadingAllAnimals[animalId, i-self._firstFrame])), int(y+2*headingLineValidationPlotLength*math.sin(trackingHeadingAllAnimals[animalId, i-self._firstFrame]))), (255,0,0), 1)
      cv2.circle(colorFrame, (midEyesPointX, midEyesPointY), 2, (255,0,255), 1)
      cv2.circle(colorFrame, (eyeX[0], eyeY[0]), 2, (0,255,255), 1)
      cv2.line(colorFrame, (eyeX[0], eyeY[0]), (int(eyeX[0]+headingLineValidationPlotLength*math.cos(eyeAngle[0])), int(eyeY[0]+headingLineValidationPlotLength*math.sin(eyeAngle[0]))), (255,0,255), 1)
      cv2.line(colorFrame, (eyeX[1], eyeY[1]), (int(eyeX[1]+headingLineValidationPlotLength*math.cos(eyeAngle[1])), int(eyeY[1]+headingLineValidationPlotLength*math.sin(eyeAngle[1]))), (255,0,255), 1)
      cv2.circle(colorFrame, (eyeX[1], eyeY[1]), 2, (0,255,255), 1)
      self._debugFrame(colorFrame, title='Eye Tracking debugging')

    # Storing the (X, Y) coordinates and angles
    trackingEyesAllAnimals[animalId, i-self._firstFrame, 0] = eyeX[0]
    trackingEyesAllAnimals[animalId, i-self._firstFrame, 1] = eyeY[0]
    trackingEyesAllAnimals[animalId, i-self._firstFrame, 2] = eyeAngle[0]
    trackingEyesAllAnimals[animalId, i-self._firstFrame, 3] = eyeArea[0]
    trackingEyesAllAnimals[animalId, i-self._firstFrame, 4] = eyeX[1]
    trackingEyesAllAnimals[animalId, i-self._firstFrame, 5] = eyeY[1]
    trackingEyesAllAnimals[animalId, i-self._firstFrame, 6] = eyeAngle[1]
    trackingEyesAllAnimals[animalId, i-self._firstFrame, 7] = eyeArea[1]

  def getAccentuateFrameForManualPointSelect(self, image):
    if self._hyperparameters["accentuateFrameForManualTailExtremityFind"]:
      frame = image.copy()
      quartileChose = 0.01
      lowVal  = int(np.quantile(frame, quartileChose))
      highVal = int(np.quantile(frame, 1 - quartileChose))
      frame[frame < lowVal]  = lowVal
      frame[frame > highVal] = highVal
      frame = frame - lowVal
      mult  = np.max(frame)
      frame = frame * (255/mult)
      frame = frame.astype(int)
      frame = (frame / np.linalg.norm(frame))*255
      return frame
    else:
      return image



import json
import matplotlib.pyplot as plt
import cv2
import numpy as np
import math
import pandas as pd
import os

from PyQt5.QtWidgets import QFileDialog


def removeLargeInstantaneousDistanceData(ZZoutputFolder, paramList):
  
  maxDistance = float(paramList[2]) if len(paramList) > 2 else 1

  pathToResultFolder =  QFileDialog.getExistingDirectory(caption='Choose the result folder for which an excel file should be generated', directory=ZZoutputFolder)
  if not pathToResultFolder:
    return

  videoName = os.path.basename(os.path.normpath(pathToResultFolder))
  
  numberOfWells       = 0
  totalNumberOfFrames = 0
  
  with open(os.path.join(pathToResultFolder, 'results_' + videoName + '.txt')) as video:
  
    supstruct = json.load(video)
    
    numberOfWells = len(supstruct['wellPoissMouv'])
    firstFrame = 0
    lastFrame  = 0
    if ('lastFrame' in supstruct) and ('firstFrame' in supstruct):
      totalNumberOfFrames = supstruct['lastFrame'] - supstruct['firstFrame'] + 1
      firstFrame = int(supstruct['firstFrame'])
      lastFrame  = int(supstruct['lastFrame'])
    else:
      raise NameError('firstFrame and lastFrame are not in the superstructure results file!')
    
    print("totalNumberOfFrames: ", totalNumberOfFrames)
    
    for numWell in range(0, len(supstruct['wellPoissMouv'])):  
      for numAnimal in range(0, len(supstruct['wellPoissMouv'][numWell])):
        for numBout in range(0, len(supstruct['wellPoissMouv'][numWell][numAnimal])):
          bout = supstruct['wellPoissMouv'][numWell][numAnimal][numBout]
          HeadX = bout['HeadX']
          HeadY = bout['HeadY']
          for i in range(0, len(HeadX)-1):
            if (HeadX[i] != float('nan') and HeadY[i] != float('nan')):
              distance = math.sqrt((HeadX[i] - HeadX[i+1])**2 + (HeadY[i] - HeadY[i+1])**2)
              if (distance > maxDistance):
                HeadX[i + 1] = float('nan')
                HeadY[i + 1] = float('nan')
          supstruct['wellPoissMouv'][numWell][numAnimal][numBout]['HeadX'] = HeadX
          supstruct['wellPoissMouv'][numWell][numAnimal][numBout]['HeadY'] = HeadY
          
  with open(os.path.join(pathToResultFolder, 'results_' + videoName + '.txt'), 'w') as outfile:
    json.dump(supstruct, outfile)



import zebrazoom
import zebrazoom.code.tracking
import zebrazoom.code.tracking.customTrackingImplementations
from zebrazoom.code.findWells import findWells
from zebrazoom.code.createSuperStruct import createSuperStruct
from zebrazoom.code.createValidationVideo import createValidationVideo
from zebrazoom.code.getHyperparameters import getHyperparameters

import pickle
import os
import shutil
import time
import cv2
import zebrazoom.videoFormatConversion.zzVideoReading as zzVideoReading
import json
import glob


class ZebraZoomVideoAnalysis:
  def __init__(self, pathToVideo, videoName, videoExt, configFile, argv, useGUI=True):
    self._pathToVideo = pathToVideo
    self._videoName = videoName
    self._configFile = configFile
    self._useGUI = useGUI
    self._videoNameWithExt = videoName + '.' + videoExt
    self.wellPositions = None
    # Getting hyperparameters
    self._hyperparameters, self._configFile = getHyperparameters(configFile, self._videoNameWithExt, os.path.join(pathToVideo, self._videoNameWithExt), argv)

    # Setting output folder
    self._outputFolderVideo = os.path.join(self._hyperparameters["outputFolder"], videoName)

  def _checkFirstAndLastFrame(self):
    # Checking first frame and last frame value
    cap   = zzVideoReading.VideoCapture(os.path.join(self._pathToVideo, self._videoNameWithExt))
    nbFrames = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
    cap.release()
    if self._hyperparameters["firstFrame"] < 0:
      print("Error for video " + self._videoName + ": The parameter 'firstFrame' in your configuration file is too small" + " (firstFrame value is " + str(self._hyperparameters["firstFrame"]) + ", number of frames in the video is " + str(nbFrames) + ")")
      raise NameError("Error for video " + self._videoName + ": The parameter 'firstFrame' in your configuration file is too small" + " (firstFrame value is " + str(self._hyperparameters["firstFrame"]) + ", number of frames in the video is " + str(nbFrames) + ")")
    if self._hyperparameters["firstFrame"] > nbFrames:
      print("Error for video " + self._videoName + ": The parameter 'firstFrame' in your configuration file is too big" + " (firstFrame value is " + str(self._hyperparameters["firstFrame"]) + ", number of frames in the video is " + str(nbFrames) + ")")
      raise NameError("Error for video " + self._videoName + ": The parameter 'firstFrame' in your configuration file is too big" + " (firstFrame value is " + str(self._hyperparameters["firstFrame"]) + ", number of frames in the video is " + str(nbFrames) + ")")
    if (self._hyperparameters["lastFrame"] < 0) or (self._hyperparameters["lastFrame"] <= self._hyperparameters["firstFrame"]):
      print("Error for video " + self._videoName + ": The parameter 'lastFrame' in your configuration file is too small" + " (lastFrame value is " + str(self._hyperparameters["lastFrame"]) + ", number of frames in the video is " + str(nbFrames) + ")")
      raise NameError("Error for video " + self._videoName + ": The parameter 'lastFrame' in your configuration file is too small" + " (lastFrame value is " + str(self._hyperparameters["lastFrame"]) + ", number of frames in the video is " + str(nbFrames) + ")")
    if self._hyperparameters["lastFrame"] > nbFrames:
      print("Warning: The parameter 'lastFrame' in your configuration file is too big so we adjusted it to the value:", nbFrames-2, "it was originally set to", self._hyperparameters["lastFrame"])
      self._hyperparameters["lastFrame"] = nbFrames - 2
      # raise NameError("Error: The parameter 'lastFrame' in your configuration file is too big")

  def _prepareOutputFolder(self):
    # Creating output folder
    if not self._hyperparameters["reloadWellPositions"] and not self._hyperparameters["reloadBackground"] and not self._hyperparameters["dontDeleteOutputFolderIfAlreadyExist"]:
      filesToKeep = {'intermediaryWellPositionReloadNoMatterWhat.txt', 'rotationAngle.txt'}
      filesToCopy = []
      if os.path.exists(self._outputFolderVideo):
        if glob.glob(os.path.join(self._outputFolderVideo, 'results_*.txt')):
          pastNumbersTaken = 1
          while os.path.exists(self._outputFolderVideo + '_PastTracking_' + str(pastNumbersTaken)) and pastNumbersTaken < 99:
            pastNumbersTaken += 1
          movedFolderName = self._outputFolderVideo + '_PastTracking_' + str(pastNumbersTaken)
          shutil.move(self._outputFolderVideo, movedFolderName)
          for filename in os.listdir(movedFolderName):
            if filename in filesToKeep:
              filesToCopy.append(os.path.join(movedFolderName, filename))
        else:
          for filename in os.listdir(self._outputFolderVideo):
            if filename not in filesToKeep:
              os.remove(os.path.join(self._outputFolderVideo, filename))
      if not os.path.exists(self._outputFolderVideo):
        if self._hyperparameters["tryCreatingFolderUntilSuccess"]:
          while True:
            try:
              os.mkdir(self._outputFolderVideo)
              break
            except OSError as e:
              print("waiting inside except")
              time.sleep(0.1)
            else:
              print("waiting")
              time.sleep(0.1)
        else:
          try:
            os.mkdir(self._outputFolderVideo)
          except OSError as e:
            time.sleep(0.1)
          else:
            time.sleep(0.1)
      for filename in filesToCopy:
        shutil.copy2(filename, self._outputFolderVideo)

  def _storeConfigUsed(self):
    '''Saving the configuration file used'''
    with open(os.path.join(self._outputFolderVideo, 'configUsed.json'), 'w') as outfile:
      json.dump(self._configFile, outfile)

  def getWellPositions(self):
    '''Get well positions'''
    if self._hyperparameters["headEmbeded"] and not self._hyperparameters["oneWellManuallyChosenTopLeft"]:
      self.wellPositions = [{"topLeftX":0, "topLeftY":0, "lengthX": self._hyperparameters["videoWidth"], "lengthY": self._hyperparameters["videoHeight"]}]
    else:
      print("start find wells")
      if self._hyperparameters["groupOfMultipleSameSizeAndShapeEquallySpacedWells"]:
        rotationAngleFile = os.path.join(self._outputFolderVideo, 'rotationAngle.txt')
        if os.path.exists(rotationAngleFile):
          with open(rotationAngleFile, 'rb') as f:
            rotationAngleParams = pickle.load(f)
          self._hyperparameters.update(rotationAngleParams)
          with open(os.path.join(self._outputFolderVideo, 'configUsed.json'), 'r') as f:
            config = json.load(f)
          config.update(rotationAngleParams)
          with open(os.path.join(self._outputFolderVideo, 'configUsed.json'), 'w') as f:
            json.dump(config, f)
      if self._hyperparameters["saveWellPositionsToBeReloadedNoMatterWhat"]:
        outfile = open(os.path.join(self._outputFolderVideo, 'intermediaryWellPositionReloadNoMatterWhat.txt'),'wb')
        self.wellPositions = findWells(os.path.join(self._pathToVideo, self._videoNameWithExt), self._hyperparameters)
        pickle.dump(self.wellPositions,outfile)
      elif os.path.exists(os.path.join(self._outputFolderVideo, 'intermediaryWellPositionReloadNoMatterWhat.txt')):
        outfile = open(os.path.join(self._outputFolderVideo, 'intermediaryWellPositionReloadNoMatterWhat.txt'), 'rb')
        self.wellPositions = pickle.load(outfile)
      elif self._hyperparameters["reloadWellPositions"]:
        outfile = open(os.path.join(self._outputFolderVideo, 'intermediaryWellPosition.txt'), 'rb')
        self.wellPositions = pickle.load(outfile)
      elif self._hyperparameters["reloadWellPositionsFromFileInZZoutputIfItExistSaveInItOtherwise"] and os.path.exists(os.path.join(self._hyperparameters["outputFolder"], 'wellPosition.txt')):
        outfile = open(os.path.join(self._hyperparameters["outputFolder"], 'wellPosition.txt'), 'rb')
        self.wellPositions = pickle.load(outfile)
      else:
        outfile = open(os.path.join(self._outputFolderVideo, 'intermediaryWellPosition.txt'),'wb')
        self.wellPositions = findWells(os.path.join(self._pathToVideo, self._videoNameWithExt), self._hyperparameters)
        pickle.dump(self.wellPositions,outfile)
        if self._hyperparameters["reloadWellPositionsFromFileInZZoutputIfItExistSaveInItOtherwise"]:
          outfile2 = open(os.path.join(self._hyperparameters["outputFolder"], 'wellPosition.txt'), 'wb')
          pickle.dump(self.wellPositions, outfile2)
          outfile2.close()
      outfile.close()
      if self._useGUI:
        from PyQt5.QtWidgets import QApplication

        app = QApplication.instance()
        if hasattr(app, "wellPositions"):
          if self.wellPositions is None:
            return
          app.wellPositions[:] = [(position['topLeftX'], position['topLeftY'], position['lengthX'], position['lengthY'])
                                  for idx, position in enumerate(self.wellPositions)]
          if self._hyperparameters["wellsAreRectangles"] or len(self._hyperparameters["oneWellManuallyChosenTopLeft"]) or int(self._hyperparameters["multipleROIsDefinedDuringExecution"]) or self._hyperparameters["noWellDetection"] or self._hyperparameters["groupOfMultipleSameSizeAndShapeEquallySpacedWells"]:
            app.wellShape = 'rectangle'
          else:
            app.wellShape = 'circle'

  def _runTracking(self):
    if 'trackingImplementation' in self._hyperparameters:
      name = self._hyperparameters['trackingImplementation']
    else:
      name = "fasterMultiprocessing" if self._hyperparameters["fasterMultiprocessing"] == 1 else "fasterMultiprocessing2" if self._hyperparameters["fasterMultiprocessing"] == 2 else "tracking"
    tracking = zebrazoom.code.tracking.get_tracking_method(name)(os.path.join(self._pathToVideo, self._videoNameWithExt), self.wellPositions, self._hyperparameters)
    if hasattr(tracking, 'useGUI'):
      tracking.useGUI = self._useGUI
    return tracking.run(), getattr(tracking, 'dataPostProcessing', None)

  def _createSuperStruct(self, paramDataPerWell):
    '''Create super structure'''
    return createSuperStruct(paramDataPerWell, self.wellPositions, self._hyperparameters, os.path.join(self._pathToVideo, self._videoNameWithExt))

  def _createValidationVideo(self, superStruct):
    '''Create validation video'''
    if not(self._hyperparameters["savePathToOriginalVideoForValidationVideo"]):
      if self._hyperparameters["copyOriginalVideoToOutputFolderForValidation"]:
        shutil.copyfile(os.path.join(self._pathToVideo, self._videoNameWithExt), os.path.join(os.path.join(self._hyperparameters["outputFolder"], self._hyperparameters["videoName"]), 'originalVideoWithoutAnyTrackingDisplayed_pleaseUseTheGUIToVisualizeTrackingPoints.avi'))
      else:
        if self._hyperparameters["createValidationVideo"]:
          infoFrame = createValidationVideo(os.path.join(self._pathToVideo, self._videoNameWithExt), superStruct, self._hyperparameters)

  def _storeResults(self, superStruct):
    path = os.path.join(os.path.join(self._hyperparameters["outputFolder"], self._hyperparameters["videoName"]), 'results_' + self._hyperparameters["videoName"] + '.txt')
    print("createSuperStruct:", path)
    with open(path, 'w') as outfile:
      json.dump(superStruct, outfile)
    if self._hyperparameters["saveSuperStructToMatlab"]:
      from scipy.io import savemat
      matlabPath = os.path.join(os.path.join(self._hyperparameters["outputFolder"], self._hyperparameters["videoName"]), 'results_' + self._hyperparameters["videoName"] + '.mat')
      videoDataResults2 = {}
      videoDataResults2['videoDataResults'] = superStruct
      savemat(matlabPath, videoDataResults2)

  def _storeVersionUsed(self):
    with open(os.path.join(os.path.join(self._hyperparameters["outputFolder"], self._hyperparameters["videoName"]), 'ZebraZoomVersionUsed.txt'), 'w') as fp:
      fp.write(zebrazoom.__version__)

  def _storeInAdditionalFolder(self):
      if os.path.isdir(self._hyperparameters["additionalOutputFolder"]):
        if self._hyperparameters["additionalOutputFolderOverwriteIfAlreadyExist"]:
          shutil.rmtree(self.hyperparameters["additionalOutputFolder"])
          while True:
            try:
              shutil.copytree(self._outputFolderVideo, self._hyperparameters["additionalOutputFolder"])
              break
            except OSError as e:
              print("waiting inside except")
              time.sleep(0.1)
            else:
              print("waiting")
              time.sleep(0.1)
        else:
          print("The path " + self._hyperparameters["additionalOutputFolder"] + " already exists. New folder not created. If you want the folder to be overwritten in such situation in future executions, set the parameter 'additionalOutputFolderOverwriteIfAlreadyExist' to 1 in your configuration file.")
      else:
        shutil.copytree(self._outputFolderVideo, self._hyperparameters["additionalOutputFolder"])

  def run(self):
    '''Run tracking'''
    # Checking that path and video exists
    if not(os.path.exists(os.path.join(self._pathToVideo, self._videoNameWithExt))):
      print("Path or video name is incorrect for", os.path.join(self._pathToVideo, self._videoNameWithExt))
      return 0

    self._checkFirstAndLastFrame()

    if self._hyperparameters["debugPauseBetweenTrackAndParamExtract"] != "justExtractParamFromPreviousTrackData":
      self._prepareOutputFolder()

    self._storeConfigUsed()

    if self._hyperparameters["trackingDL"]:
      from torch.multiprocessing import Process
    else:
      from multiprocessing import Process

    # Launching GUI algoFollower if necessary
    if self._hyperparameters["popUpAlgoFollow"]:
      import zebrazoom.code.popUpAlgoFollow as popUpAlgoFollow

      popUpAlgoFollow.createTraceFile("starting ZebraZoom analysis on " + self._videoName)
      p = Process(target=popUpAlgoFollow.initialise)
      p.start()

    self.getWellPositions()
    if int(self._hyperparameters["exitAfterWellsDetection"]):
      print("exitAfterWellsDetection")
      if self._hyperparameters["popUpAlgoFollow"]:
        import zebrazoom.code.popUpAlgoFollow as popUpAlgoFollow

        popUpAlgoFollow.prepend("ZebraZoom Analysis finished for " + self._videoName)
      raise ValueError

    paramDataPerWell, postProcessingCb = self._runTracking()

    if self._hyperparameters["debugPauseBetweenTrackAndParamExtract"] != "justSaveTrackData":
      superStruct = self._createSuperStruct(paramDataPerWell)
      self._createValidationVideo(superStruct)
      if postProcessingCb is not None:
        superStruct = postProcessingCb(self._outputFolderVideo, superStruct)
      self._storeResults(superStruct)

    self._storeVersionUsed()

    # Copying output result folder in another folder
    if len(self._hyperparameters["additionalOutputFolder"]):
      self._storeInAdditionalFolder()

    if self._hyperparameters["popUpAlgoFollow"]:
      import zebrazoom.code.popUpAlgoFollow as popUpAlgoFollow

      popUpAlgoFollow.prepend("ZebraZoom Analysis finished for " + self._videoName)
      # popUpAlgoFollow.prepend("")
      # if self._hyperparameters["closePopUpWindowAtTheEnd"]:
        # popUpAlgoFollow.prepend("ZebraZoom Analysis all finished")



import math
import os
import sys

import cv2
import json
import numpy as np

from PyQt5.QtCore import Qt, QSize, QTimer
from PyQt5.QtWidgets import QApplication, QCheckBox, QFileDialog, QMessageBox, QLabel, QSlider, QVBoxLayout

import zebrazoom.code.paths as paths
import zebrazoom.code.util as util
import zebrazoom.videoFormatConversion.zzVideoReading as zzVideoReading
from zebrazoom.code.createValidationVideo import calculateInfoFrame, drawInfoFrame
from zebrazoom.code.getHyperparameters import getHyperparametersSimple
from zebrazoom.code.preprocessImage import preprocessImage


def readValidationVideo(videoPath, folderName, configFilePath, numWell, numAnimal, zoom, start, framesToShow=0, ZZoutputLocation=''):
  s1  = "ZZoutput"
  s2  = folderName
  s3b = "results_"
  s4  = folderName
  s5  = ".avi"
  s5b = ".txt"

  if len(ZZoutputLocation):
    initialPath = ZZoutputLocation
  else:
    initialPath = paths.getDefaultZZoutputFolder()

  with open(os.path.join(initialPath, os.path.join(s2, 'configUsed.json'))) as f:
    configTemp = json.load(f)
  hyperparameters = getHyperparametersSimple(configTemp)
  
  resultsPath = os.path.join(initialPath, os.path.join(s2, s3b + s4 + s5b))

  if not(os.path.exists(resultsPath)):
    mypath = os.path.join(initialPath, s2)
    onlyfiles = [f for f in os.listdir(mypath) if os.path.isfile(os.path.join(mypath, f))]
    resultFile = ''
    for fileName in onlyfiles:
      if 'results_' in fileName:
        resultFile = fileName
    resultsPath = os.path.join(initialPath, os.path.join(s2, resultFile))

  with open(resultsPath) as f:
    supstruct = json.load(f)
  
  if hyperparameters["savePathToOriginalVideoForValidationVideo"]:
    videoPath = supstruct["pathToOriginalVideo"]
    if not os.path.exists(videoPath):
      app = QApplication.instance()
      if QMessageBox.critical(app.window, "Video not found", "Cannot display the validation video because the video used to run the tracking can no longer be found. Would you like to update the video path stored in the results file?",
                              buttons=QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No, defaultButton=QMessageBox.StandardButton.Yes) != QMessageBox.StandardButton.Yes:
        return
      videoName, _ = QFileDialog.getOpenFileName(app.window, 'Select video', os.path.expanduser("~"))
      videoPath = supstruct["pathToOriginalVideo"] = videoName
      with open(resultsPath, 'w') as f:
        json.dump(supstruct, f)
  else:
    if hyperparameters["copyOriginalVideoToOutputFolderForValidation"] and os.path.exists(os.path.join(initialPath, os.path.join(s1, os.path.join(s2, 'originalVideoWithoutAnyTrackingDisplayed_pleaseUseTheGUIToVisualizeTrackingPoints.avi')))):
      # The "exist" check above is only to insure compatibility with videos tracked prior to this update
      videoPath = os.path.join(initialPath, os.path.join(s2, 'originalVideoWithoutAnyTrackingDisplayed_pleaseUseTheGUIToVisualizeTrackingPoints.avi'))
    else:
      videoPath = os.path.join(initialPath, os.path.join(s2, s4 + s5))

  if not(os.path.exists(videoPath)):
    mypath = os.path.join(initialPath, s2)
    onlyfiles = [f for f in os.listdir(mypath) if os.path.isfile(os.path.join(mypath, f))]
    resultFile = ''
    for fileName in onlyfiles:
      if '.avi' in fileName:
        resultFile = fileName
    videoPath = os.path.join(initialPath, os.path.join(s2, resultFile))

  cap = zzVideoReading.VideoCapture(videoPath)

  nx    = int(cap.get(3))
  ny    = int(cap.get(4))
  max_l = int(cap.get(7))
  sliderRange = (supstruct["firstFrame"], supstruct["lastFrame"] - 1) if hyperparameters["copyOriginalVideoToOutputFolderForValidation"] or hyperparameters["savePathToOriginalVideoForValidationVideo"] else (0, max_l -1)

  if not("firstFrame" in supstruct):
    supstruct["firstFrame"] = 1
    print("supstruct['firstFrame'] not found")

  infoWells = []

  HeadX = np.zeros(max_l + supstruct["firstFrame"])
  HeadY = np.zeros(max_l + supstruct["firstFrame"])

  if ((numWell != -1) and (zoom)):
    lastEnd = 0
    lastXpos = supstruct["wellPoissMouv"][numWell][numAnimal][0]["HeadX"][0]
    lastYpos = supstruct["wellPoissMouv"][numWell][numAnimal][0]["HeadY"][0]
    for k in range(0,len(supstruct["wellPoissMouv"][numWell][numAnimal])):
      beg = supstruct["wellPoissMouv"][numWell][numAnimal][k]["BoutStart"]
      end = supstruct["wellPoissMouv"][numWell][numAnimal][k]["BoutEnd"]
      for l in range(lastEnd, beg):
        HeadX[l] = lastXpos
        HeadY[l] = lastYpos
      for l in range(beg, end):
        HeadX[l]  = supstruct["wellPoissMouv"][numWell][numAnimal][k]["HeadX"][l-beg]
        HeadY[l]  = supstruct["wellPoissMouv"][numWell][numAnimal][k]["HeadY"][l-beg]
      lastEnd = end
      lastXpos = supstruct["wellPoissMouv"][numWell][numAnimal][k]["HeadX"][end-1-beg]
      lastYpos = supstruct["wellPoissMouv"][numWell][numAnimal][k]["HeadY"][end-1-beg]

    for l in range(lastEnd, max_l + supstruct["firstFrame"]):
      HeadX[l] = lastXpos
      HeadY[l] = lastYpos

  # /* Getting the info about well positions */
  analyzeAllWellsAtTheSameTime = 0
  if (analyzeAllWellsAtTheSameTime == 0):
    for i in range(0, len(supstruct["wellPositions"])):
      x = 0
      y = 0
      lengthX = 0
      lengthY = 0
      rectangleWellArea = 1
      if (rectangleWellArea == 0): # circular wells
        x = supstruct["wellPositions"][i]["topLeftX"]
        y = supstruct["wellPositions"][i]["topLeftY"]
        r = supstruct["wellPositions"][i]["diameter"]
        lengthX = 300 # wellOutputVideoDiameter;
        lengthY = 300 # wellOutputVideoDiameter;
      else:
        x = supstruct["wellPositions"][i]["topLeftX"]
        y = supstruct["wellPositions"][i]["topLeftY"]
        lengthX = supstruct["wellPositions"][i]["lengthX"]
        lengthY = supstruct["wellPositions"][i]["lengthY"]
      if (x < 0):
        x = 0
      if (y < 0):
        y = 0
      infoWells.append([x, y, lengthX, lengthY])
  else:
    infoWells.append([0, 0, nx, ny])
  
  infoFrame = None
  if hyperparameters["copyOriginalVideoToOutputFolderForValidation"] or hyperparameters["savePathToOriginalVideoForValidationVideo"]:
    infoFrame, colorModifTab = calculateInfoFrame(supstruct, hyperparameters, max_l)
  
  x = 0
  y = 0
  lengthX = 0
  lengthY = 0
  if (numWell != -1):
    x = infoWells[numWell][0]
    y = infoWells[numWell][1]
    lengthX = infoWells[numWell][2]
    lengthY = infoWells[numWell][3]
  else:
    lengthX = nx
    lengthY = ny

  l = start - supstruct["firstFrame"] + 1 if start > 0 else 0
  if hyperparameters["copyOriginalVideoToOutputFolderForValidation"] or hyperparameters["savePathToOriginalVideoForValidationVideo"]:
    l += supstruct["firstFrame"]

  xOriginal = x
  yOriginal = y

  def getFrame():
    nonlocal x
    nonlocal y
    nonlocal lengthX
    nonlocal lengthY

    l = frameSlider.value()
    if timer.isActive() and (l == frameSlider.maximum() or stopTimer):
      timer.stop()

    cap.set(1, l)
    ret, img = cap.read()
    
    if hyperparameters["imagePreProcessMethod"]:
      img = preprocessImage(img, hyperparameters)
    
    if infoFrame is not None and plotTrackingPointsCheckbox.isChecked():
      drawInfoFrame(l, img, infoFrame, colorModifTab, hyperparameters)

    if numWell != -1 and zoom:
      length = 250
      xmin = int(HeadX[l + supstruct["firstFrame"] - 1] - length/2)
      xmax = int(HeadX[l + supstruct["firstFrame"] - 1] + length/2)
      ymin = int(HeadY[l + supstruct["firstFrame"] - 1] - length/2)
      ymax = int(HeadY[l + supstruct["firstFrame"] - 1] + length/2)

      x = max(xmin + xOriginal, 0)
      y = max(ymin + yOriginal, 0)
      lengthX = xmax - xmin
      lengthY = ymax - ymin

      if y + lengthY >= len(img):
        lengthY = len(img) - y - 1
      if x + lengthX >= len(img[0]):
        lengthX = len(img[0]) - x - 1

    if (numWell != -1):
      img = img[y:y+lengthY, x:x+lengthX]

    frameNumber = l if hyperparameters["copyOriginalVideoToOutputFolderForValidation"] or hyperparameters["savePathToOriginalVideoForValidationVideo"] else (l + supstruct["firstFrame"])
    if lengthX > 100 and lengthY > 100:
      font = cv2.FONT_HERSHEY_SIMPLEX
      cv2.putText(img,str(frameNumber),(int(lengthX-110), int(lengthY-30)),font,1,(0,255,0))
    else:
      blank_image = np.zeros((len(img)+30, len(img[0]), 3), np.uint8)
      blank_image[0:len(img), 0:len(img[0])] = img
      img = blank_image
      font = cv2.FONT_HERSHEY_SIMPLEX
      cv2.putText(img, str(frameNumber), (int(0), int(lengthY+25)), font, 1, (0,255,0))

    return img

  layout = QVBoxLayout()

  video = QLabel()
  layout.addWidget(video, stretch=1)

  frameSlider = QSlider(Qt.Orientation.Horizontal)
  frameSlider.setFocusPolicy(Qt.FocusPolicy.StrongFocus)
  frameSlider.setPageStep(50)
  frameSlider.setRange(*sliderRange)
  frameSlider.setValue(l)
  frameSlider.valueChanged.connect(lambda: util.setPixmapFromCv(getFrame(), video))
  layout.addWidget(frameSlider)
  shortcutsLabel = QLabel("Left Arrow, Right Arrow, Page Up, Page Down, Home and End keys can be used to navigate through the video.")
  shortcutsLabel.setWordWrap(True)
  layout.addWidget(shortcutsLabel)
  if infoFrame is not None:
    plotTrackingPointsCheckbox = QCheckBox("Display tracking points")
    plotTrackingPointsCheckbox.setChecked(True)
    plotTrackingPointsCheckbox.toggled.connect(lambda: timer.isActive() or util.setPixmapFromCv(getFrame(), video))
    layout.addWidget(plotTrackingPointsCheckbox)

  stopTimer = True
  timer = QTimer()
  timer.setInterval(1)

  def nextFrame():
    nonlocal stopTimer
    stopTimer = False
    frameSlider.setValue(frameSlider.value() + 1)
    stopTimer = True
  timer.timeout.connect(nextFrame)

  startFrame = getFrame()
  timer.start()
  util.showDialog(layout, title="Video", labelInfo=(startFrame, video))
  timer.stop()
  cap.release()



import numpy as np
import cv2
import zebrazoom.videoFormatConversion.zzVideoReading as zzVideoReading
import math
import os
import pickle


class TailTrackingDifficultBackgroundMixin:
  def _getCoordinates(self, frame, title, zoomable, dialog):
    raise ValueError("Some of the required inputs were not specified, please specify them in advance or run tracking from the GUI.") # TODO: a better exception?

  @staticmethod
  def __simpleOptimalValueSearch(PtClosest, contour, unitVector, lenX, lenY):
    factor = 0
    dist = 1
    maxDist = 0
    indMax = 0
    testCenter = PtClosest + factor * unitVector
    while (dist > 0) and (factor < 20) and (testCenter[0] >= 0) and (testCenter[1] >= 0) and (testCenter[0] < lenX) and (testCenter[1] < lenY):
      factor = factor + 1
      testCenter = PtClosest + factor * unitVector
      testCenter = testCenter.astype(int)
      dist = cv2.pointPolygonTest(contour, (testCenter[0], testCenter[1]), True)
      if dist > maxDist:
        maxDist = dist
        indMax  = factor

    testCenter = PtClosest + indMax * unitVector
    testCenter = testCenter.astype(int)

    return testCenter

  def __reajustCenterOfMassIfNecessary(self, contour, x, y, lenX, lenY):
    inside = cv2.pointPolygonTest(contour, (x, y), True)
    if inside < 0:

      minDist = 100000000000000
      indMin  = 0
      for i in range(0, len(contour)):
        Pt = contour[i][0]
        dist = math.sqrt((Pt[0] - x)**2 + (Pt[1] - y)**2)
        if dist < minDist:
          minDist = dist
          indMin  = i
      PtClosest = contour[indMin][0]
      unitVector = np.array([PtClosest[0] - x, PtClosest[1] - y])
      unitVectorLength = math.sqrt(unitVector[0]**2 + unitVector[1]**2)
      unitVector[0] = unitVector[0] / unitVectorLength
      unitVector[1] = unitVector[1] / unitVectorLength
      if False:
        factor = 5
        testCenter = PtClosest + factor * unitVector
        testCenter = testCenter.astype(int)
        while (cv2.pointPolygonTest(contour, (testCenter[0], testCenter[1]), True) <= 0) and (factor > 1):
          factor = factor - 1
          testCenter = PtClosest + factor * unitVector
      else:
        testCenter = self.__simpleOptimalValueSearch(PtClosest, contour, unitVector, lenX, lenY)

      x = testCenter[0]
      y = testCenter[1]

    return [x, y]

  @staticmethod
  def __fillWhiteHoles(frame):
    frameBeforeWhiteFill = frame.copy()
    im_floodfill = frame.copy()
    h, w = frame.shape[:2]
    mask = np.zeros((h+2, w+2), np.uint8)
    cv2.floodFill(im_floodfill, mask, (0,0), 255);
    im_floodfill_inv = cv2.bitwise_not(im_floodfill)
    frame = frame | im_floodfill_inv

    if cv2.countNonZero(frame) > (len(frame) * len(frame[0])) * 0.95:
      frame = frameBeforeWhiteFill
      print("BACK TO IMAGE BEFORE WHITE FILL")

    return frame

  @staticmethod
  def __erodeThenAddWhiteBorders(frame, kernel):
    frame = cv2.erode(frame, kernel, iterations=1)
    frame[0,:] = 255
    frame[len(frame)-1,:] = 255
    frame[:,0] = 255
    frame[:,len(frame[0])-1] = 255
    return frame

  @staticmethod
  def __erodeThenDilateThenAddWhiteBorders(frame, kernel, nbOfIterations):
    frame = cv2.erode(frame, kernel, iterations=nbOfIterations)
    frame = cv2.dilate(frame, kernel, iterations=nbOfIterations)
    frame[0,:] = 255
    frame[len(frame)-1,:] = 255
    frame[:,0] = 255
    frame[:,len(frame[0])-1] = 255
    return frame

  @staticmethod
  def __addWhiteBorders(frame):
    frame[0,:] = 255
    frame[len(frame)-1,:] = 255
    frame[:,0] = 255
    frame[:,len(frame[0])-1] = 255
    return frame

  def _fishTailTrackingDifficultBackground(self, wellNumber):
    chooseValueForAnimalBodyArea = -1 #600 # -1
    showInitialVideo = True # This is not used anymore, should be removed soon
    iterativelyErodeEachImage = False
    showFramesForDebugging = self._hyperparameters["debugTracking"]
    dist2Threshold = 400
    historyLength  = 1000
    reduceImageResolutionPercentage = self._hyperparameters["reduceImageResolutionPercentage"]

    pathToVideo = os.path.split(self._videoPath)[0]

    kernel  = np.ones((3, 3), np.uint8)
    ROIHalfDiam = -1

    class CustomError(Exception):
      pass

    cap = zzVideoReading.VideoCapture(self._videoPath)
    if (cap.isOpened()== False):
      print("Error opening video stream or file")
    frame_width  = int(cap.get(3))
    frame_height = int(cap.get(4))

    fgbg = cv2.createBackgroundSubtractorKNN(dist2Threshold = dist2Threshold, history = historyLength)

    if os.path.exists(os.path.join(pathToVideo, self._videoName + '_BackgroundKNN_' + str(historyLength) + '.pkl')):
      print("Background already pre-calculated")
      with open(os.path.join(pathToVideo, self._videoName + '_BackgroundKNN_' + str(historyLength) + '.pkl'), 'rb') as handle:
        listOfFramesToSave = pickle.load(handle)
      for i in range(0, len(listOfFramesToSave)):
        print("calculating background, currently at frame:", i)
        frame = listOfFramesToSave[i]
        fgmask = fgbg.apply(frame)
    else:
      print("Background calculation starting")

      listOfFramesToSave = []
      framesToKeep = [i for i in range(0, self._lastFrame - 1, int((self._lastFrame - 1) / historyLength if (self._lastFrame - 1) / historyLength >= 1 else 1))]
      cap.set(1, 0)
      for i in range(0, self._lastFrame):
        ret, frame = cap.read()
        if i in framesToKeep and ret:
          print(i, self._lastFrame)
          frame = cv2.resize(frame, (int(frame_width * reduceImageResolutionPercentage), int(frame_height * reduceImageResolutionPercentage)), interpolation = cv2.INTER_AREA)
          listOfFramesToSave.append(frame)

      listOfFramesToSave.reverse()

      for i in range(0, len(listOfFramesToSave)):
        print("calculating background, currently at frame:", i)
        frame = listOfFramesToSave[i]
        fgmask = fgbg.apply(frame)

      # listOfFramesToSave = []
      # for i in range(0, self._lastFrame - 1, int((self._lastFrame - 1) / historyLength if (self._lastFrame - 1) / historyLength >= 1 else 1))]
        # cap.set(1, self._lastFrame - 1 - i)
        # print("calculating background, currently at frame:", self._lastFrame - 1 - i)
        # ret, frame = cap.read()
        # if ret:
          # frame = cv2.resize(frame, (int(frame_width * reduceImageResolutionPercentage), int(frame_height * reduceImageResolutionPercentage)), interpolation = cv2.INTER_AREA)
          # fgmask = fgbg.apply(frame)
          # listOfFramesToSave.append(frame)
      cap.release()
      with open(os.path.join(pathToVideo, self._videoName + '_BackgroundKNN_' + str(historyLength) + '.pkl'), 'wb') as handle:
        pickle.dump(listOfFramesToSave, handle) #, protocol=pickle.HIGHEST_PROTOCOL)

    cap = zzVideoReading.VideoCapture(self._videoPath)

    i = self._firstFrame
    cap.set(1, self._firstFrame)

    while i < self._lastFrame:

      print(i, self._lastFrame)

      ret, frame = cap.read()

      if ret:

        frameInitialImage = frame.copy()

        if self._hyperparameters["reduceImageResolutionPercentage"]:
          frame = cv2.resize(frame, (int(frame_width * self._hyperparameters["reduceImageResolutionPercentage"]), int(frame_height * self._hyperparameters["reduceImageResolutionPercentage"])), interpolation = cv2.INTER_AREA)
          frameInitialImage = frame.copy()

        if i == self._firstFrame:

          previousCenterDetectedX, previousCenterDetectedY = self._getCoordinates(frame, "Click on the center of the head of the animal", True, True)
          tailTipX, tailTipY = self._getCoordinates(frame, "Click on the tip of the tail of the same animal", True, True)

          frame2 = frame.copy()
          frame2 = fgbg.apply(frame2)
          ret, frame2 = cv2.threshold(frame2, 0, 255, cv2.THRESH_BINARY)
          frame2 = self.__fillWhiteHoles(frame2)
          frame2 = 255 - frame2
          if iterativelyErodeEachImage:
            frame2 = self.__erodeThenAddWhiteBorders(frame2, kernel)
          else:
            frame2 = self.__addWhiteBorders(frame2)
          contours, hierarchy = cv2.findContours(frame2, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
          for contour in contours:
            if cv2.pointPolygonTest(contour, (previousCenterDetectedX, previousCenterDetectedY), False) >= 0:
              animalBodyArea = cv2.contourArea(contour)
              ROIHalfDiam = int(math.sqrt(animalBodyArea) * 4)
          print("animalBodyArea:", animalBodyArea)
          print("ROIHalfDiam:", ROIHalfDiam)
          if showFramesForDebugging:
            self._debugFrame(frame2, title='frame0')

        if chooseValueForAnimalBodyArea > 0:
          animalBodyArea = chooseValueForAnimalBodyArea

        if ROIHalfDiam == -1:
          raise CustomError("An error occurred")

        diffAreaAuthorized = 0.5
        self._hyperparameters["minAreaBody"] = int(animalBodyArea - animalBodyArea * diffAreaAuthorized)
        self._hyperparameters["maxAreaBody"] = int(animalBodyArea + animalBodyArea * diffAreaAuthorized)
        self._hyperparameters["minArea"] = int(animalBodyArea - animalBodyArea * diffAreaAuthorized)
        self._hyperparameters["maxArea"] = int(animalBodyArea + animalBodyArea * diffAreaAuthorized)
        self._hyperparameters["headSize"]    = int(math.sqrt(animalBodyArea) * 2)
        self._hyperparameters["minTailSize"] = int(math.sqrt(animalBodyArea) * 0.5)
        self._hyperparameters["maxTailSize"] = int(math.sqrt(animalBodyArea) * 4)

        xmin = previousCenterDetectedX - ROIHalfDiam if previousCenterDetectedX - ROIHalfDiam > 0 else 0
        xmax = previousCenterDetectedX + ROIHalfDiam if previousCenterDetectedX + ROIHalfDiam < len(frame[0]) else len(frame[0]) - 1
        ymin = previousCenterDetectedY - ROIHalfDiam if previousCenterDetectedY - ROIHalfDiam > 0 else 0
        ymax = previousCenterDetectedY + ROIHalfDiam if previousCenterDetectedY + ROIHalfDiam < len(frame) else len(frame) - 1

        previousCenterDetectedXROICoordinates = previousCenterDetectedX - xmin
        previousCenterDetectedYROICoordinates = previousCenterDetectedY - ymin

        if showInitialVideo:
          initialImage = frame.copy()

        frame = fgbg.apply(frame)

        frame = frame[ymin:ymax, xmin:xmax]

        ret, frame = cv2.threshold(frame, 0, 255, cv2.THRESH_BINARY)

        frame = self.__fillWhiteHoles(frame)

        frame = 255 - frame

        countNbOfRightArea = 0
        distanceToPreviousCenterDetected = 100000000000000000000
        countNbTries = 0
        newCenterDetectedX_ROICordinates = previousCenterDetectedXROICoordinates
        newCenterDetectedY_ROICordinates = previousCenterDetectedYROICoordinates
        mostLikelyContour = 0
        while countNbOfRightArea == 0 and countNbTries < 10:
          countNbTries = countNbTries + 1
          if iterativelyErodeEachImage:
            frame = self.__erodeThenAddWhiteBorders(frame, kernel)
          else:
            frame = self.__addWhiteBorders(frame)
          if showFramesForDebugging:
            self._debugFrame(frame, title='frame1')

          nbOfIterations = 1
          while cv2.countNonZero(255 - frame) < animalBodyArea and nbOfIterations < 10:
            print("try " + str(nbOfIterations))
            frame = self.__erodeThenDilateThenAddWhiteBorders(frame, kernel, nbOfIterations)
            nbOfIterations += 1
            if showFramesForDebugging:
              self._debugFrame(frame, title='frame%d' % str(nbOfIterations))

          if nbOfIterations > 1:
            dist2t = fgbg.getDist2Threshold()
            fgbg.setDist2Threshold(dist2t - dist2t * 0.1)
            print("dist2t set to a smaller value:", dist2t - dist2t * 0.1)
          else:
            if cv2.countNonZero(255 - frame) > 1.5 * animalBodyArea:
              dist2t = fgbg.getDist2Threshold()
              fgbg.setDist2Threshold(dist2t + dist2t * 0.1)
              print("dist2t set to a bigger value:", dist2t + dist2t * 0.1)

          contours, hierarchy = cv2.findContours(frame, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
          for contour in contours:
            contourArea = cv2.contourArea(contour)
            if contourArea > self._hyperparameters["minAreaBody"] and contourArea < self._hyperparameters["maxAreaBody"]:
              countNbOfRightArea = countNbOfRightArea + 1
              M = cv2.moments(contour)
              cx = int(M['m10']/M['m00'])
              cy = int(M['m01']/M['m00'])
              [cx, cy] = self.__reajustCenterOfMassIfNecessary(contour, cx, cy, len(frame[0]), len(frame))
              if math.sqrt((previousCenterDetectedXROICoordinates - cx)**2 + (previousCenterDetectedYROICoordinates - cy)**2) < distanceToPreviousCenterDetected:
                newCenterDetectedX_ROICordinates = cx
                newCenterDetectedY_ROICordinates = cy
                mostLikelyContour = contour

          if not(iterativelyErodeEachImage):
            countNbTries = 10
          # print("First: newCenterDetectedX_ROICordinates, newCenterDetectedY_ROICordinates:", newCenterDetectedX_ROICordinates, newCenterDetectedY_ROICordinates)

          if type(mostLikelyContour) != int:

            cntXmin = 100000000000000
            cntYmin = 100000000000000
            cntXmax = -1
            cntYmax = -1

            for point in mostLikelyContour:
              if point[0][0] < cntXmin:
                cntXmin = point[0][0]
              if point[0][1] < cntYmin:
                cntYmin = point[0][1]
              if point[0][0] > cntXmax:
                cntXmax = point[0][0]
              if point[0][1] > cntYmax:
                cntYmax = point[0][1]

              cntXmin = cntXmin - 10
              cntYmin = cntYmin - 10
              cntXmax = cntXmax + 10
              cntYmax = cntYmax + 10

            if False:

              print("This is False")
              # for point in mostLikelyContour:
                # point[0][0] = point[0][0] - cntXmin
                # point[0][1] = point[0][1] - cntYmin

              # blank = np.zeros((cntYmax - cntYmin, cntXmax - cntXmin, 3), np.uint8)
              # blank = 255 - blank

              # blank = cv2.fillPoly(blank, pts =[mostLikelyContour], color=(0, 0, 0))

              # blank = cv2.dilate(blank, kernel, iterations=1)
              # blankIni = blank.copy()

              # nbDilationCount = 0
              # blankPreviousIteration = blank

              # while cv2.countNonZero(cv2.cvtColor(255 - blank, cv2.COLOR_BGR2GRAY)) != 0 and nbDilationCount < 10:
                # blankPreviousIteration = cv2.cvtColor(blank.copy(), cv2.COLOR_BGR2GRAY)
                # blank = cv2.dilate(blank, kernel, iterations=1)
                # nbDilationCount = nbDilationCount + 1

              # contours, hierarchy = cv2.findContours(blankPreviousIteration, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
              # largestContourArea = 0
              # for contour in contours:
                # contourArea = cv2.contourArea(contour)
                # if contourArea > largestContourArea:
                  # largestContourArea = contourArea
                  # M = cv2.moments(contour)
                  # x = int(M['m10']/M['m00'])
                  # y = int(M['m01']/M['m00'])
                  # if False:
                    # [newCenterDetectedX_ROICordinates, newCenterDetectedY_ROICordinates] = self.__reajustCenterOfMassIfNecessary(contour, x, y, len(blankPreviousIteration[0]), len(blankPreviousIteration))
                  # newCenterDetectedX_ROICordinates = x
                  # newCenterDetectedY_ROICordinates = y

            else:

              cntXmin = 10
              cntYmin = 10

              blank = np.zeros((xmax - xmin, ymax - ymin, 3), np.uint8)
              blank = 255 - blank

              blank = cv2.fillPoly(blank, pts =[mostLikelyContour], color=(0, 0, 0))

              blank = cv2.dilate(blank, kernel, iterations=1)
              blankIni = blank.copy()

              nbDilationCount = 0
              blankPreviousIteration = blank

              while cv2.countNonZero(cv2.cvtColor(255 - blank, cv2.COLOR_BGR2GRAY)) != 0 and nbDilationCount < 10:
                blankPreviousIteration = cv2.cvtColor(blank.copy(), cv2.COLOR_BGR2GRAY)
                blank = cv2.dilate(blank, kernel, iterations=1)
                nbDilationCount = nbDilationCount + 1

              if type(blankPreviousIteration) == np.ndarray and len(blankPreviousIteration) and len(blankPreviousIteration[0]) and type(blankPreviousIteration[0][0]) == np.uint8:
                contours, hierarchy = cv2.findContours(blankPreviousIteration, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
                largestContourArea = 0
                for contour in contours:
                  contourArea = cv2.contourArea(contour)
                  if contourArea > largestContourArea and contourArea < (len(blankPreviousIteration)*len(blankPreviousIteration[0])) * 0.8:
                    largestContourArea = contourArea
                    M = cv2.moments(contour)
                    x = int(M['m10']/M['m00'])
                    y = int(M['m01']/M['m00'])
                    # if False:
                      # [newCenterDetectedX_ROICordinates, newCenterDetectedY_ROICordinates] = self.__reajustCenterOfMassIfNecessary(contour, x, y, len(blankPreviousIteration[0]), len(blankPreviousIteration))
                    newCenterDetectedX_ROICordinates = x
                    newCenterDetectedY_ROICordinates = y

        # print("Second: newCenterDetectedX_ROICordinates, newCenterDetectedY_ROICordinates:", newCenterDetectedX_ROICordinates, newCenterDetectedY_ROICordinates)

        self._trackingHeadTailAllAnimals[0, i-self._firstFrame][0][0] = newCenterDetectedX_ROICordinates
        self._trackingHeadTailAllAnimals[0, i-self._firstFrame][0][1] = newCenterDetectedY_ROICordinates

        if self._hyperparameters["trackTail"] == 1:
          if type(blankIni) == np.ndarray and len(blankIni) and len(blankIni[0]):
            if type(blankIni[0][0]) != np.uint8:
              blankIni = cv2.cvtColor(blankIni, cv2.COLOR_BGR2GRAY)
            for animalId in range(0, self._hyperparameters["nbAnimalsPerWell"]):
              self._tailTracking(animalId, i, blankIni, blankIni, blankIni, 0, self._trackingHeadTailAllAnimals, self._trackingHeadingAllAnimals, 0, 0, 0, blankIni, 0, wellNumber)
          else:
            print("problem")

        for j in range(0, len(self._trackingHeadTailAllAnimals[0, i-self._firstFrame])):
          self._trackingHeadTailAllAnimals[0, i-self._firstFrame][j][0] += xmin + cntXmin - 10
          self._trackingHeadTailAllAnimals[0, i-self._firstFrame][j][1] += ymin + cntYmin - 10

        self._debugTracking(i, self._trackingHeadTailAllAnimals, self._trackingHeadingAllAnimals, initialImage)
        # if showInitialVideo:
          # initialImage = cv2.circle(initialImage, (newCenterDetectedX_ROICordinates + xmin + cntXmin - 10, newCenterDetectedY_ROICordinates + ymin + cntYmin - 10), 3, (255, 0, 0), 2)
          # cv2.imshow("frame", initialImage)
          # cv2.waitKey(0)
        # else:
          # frame = cv2.circle(frame, (newCenterDetectedX_ROICordinates, newCenterDetectedY_ROICordinates), 3, (255, 0, 0), 2)
          # cv2.imshow("frame", frame)
          # cv2.waitKey(0)

        previousCenterDetectedX = newCenterDetectedX_ROICordinates + xmin + cntXmin - 10
        previousCenterDetectedY = newCenterDetectedY_ROICordinates + ymin + cntYmin - 10

      i = i + 1

    return [self._trackingHeadTailAllAnimals, self._trackingHeadingAllAnimals, 0]



import os
import numpy as np
import torch
import torch.utils.data
from PIL import Image
import time

import torchvision
from torchvision.models.detection.faster_rcnn import FastRCNNPredictor
from torchvision.models.detection.mask_rcnn import MaskRCNNPredictor

from   zebrazoom.code.deepLearningFunctions.PyTorchFunctions.engine     import train_one_epoch, evaluate
import zebrazoom.code.deepLearningFunctions.PyTorchFunctions.utils      as utils
import zebrazoom.code.deepLearningFunctions.PyTorchFunctions.transforms as T


def get_instance_segmentation_model(num_classes):

  model = torchvision.models.detection.maskrcnn_resnet50_fpn(pretrained=True)
  in_features = model.roi_heads.box_predictor.cls_score.in_features
  model.roi_heads.box_predictor = FastRCNNPredictor(in_features, num_classes)
  in_features_mask = model.roi_heads.mask_predictor.conv5_mask.in_channels
  hidden_layer = 256
  model.roi_heads.mask_predictor = MaskRCNNPredictor(in_features_mask, hidden_layer, num_classes)
  
  return model


def loadDLmodel(pathToSavedModel, isUnet):
  
  if isUnet:
    from zebrazoom.code.deepLearningFunctions.unetModel.unet_model import UNet
    model = UNet(n_channels=1, n_classes=2, bilinear=1)
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    model.to(device=device)
    state_dict = torch.load(pathToSavedModel, map_location=device)
    mask_values = state_dict.pop('mask_values', [0, 1]) # Should probably delete this line
    model.load_state_dict(state_dict)
    model.eval()
  else:
    num_classes = 2
    device = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')
    model = get_instance_segmentation_model(num_classes)
    model.to(device)
    model.load_state_dict(torch.load(pathToSavedModel, map_location=device))
    model.eval()
  
  return model

